# Copyright 2014 Flax & Teal Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Abstraction layer for task operations.

Access dask, or other, backends consistently using this module. It provides
decorators and execution calls that manage tasks. Note that the `task`
decorator should be called with no arguments, and will return the appropriate
decorator for the current backend.

Typical usage example:

```python
>>> @task()
... def increment(num: int) -> int:
...     return num + 1

```
"""

import inspect
import importlib
import sys
from enum import Enum
from functools import cached_property
from collections.abc import Callable
from typing import Any, ParamSpec, TypeVar, cast, Generator, Unpack, Literal
from types import TracebackType
from attrs import has as attrs_has
from dataclasses import is_dataclass
import traceback
from concurrent.futures import ThreadPoolExecutor
from contextvars import ContextVar, copy_context
from contextlib import contextmanager

from .utils import is_firm, make_traceback, is_expr
from .workflow import (
    expr_to_references,
    unify_workflows,
    UNSET,
    Workflow,
    Lazy,
    LazyEvaluation,
    Target,
    LazyFactory,
    Parameter,
    ParameterReference,
    param,
    Task,
    is_task,
)
from .backends._base import BackendModule
from .annotations import FunctionAnalyser
from .core import (
    get_configuration,
    set_configuration,
    IteratedGenerator,
    ConstructConfigurationTypedDict,
    Reference,
)

Param = ParamSpec("Param")
RetType = TypeVar("RetType")
T = TypeVar("T")


class Backend(Enum):
    """Stringy enum representing available backends."""

    DASK = "dask"


DEFAULT_BACKEND = Backend.DASK


class TaskManager:
    """Overarching backend-agnostic task manager.

    Gatekeeps the specific backend implementation. This can be
    instantiated without choosing a backend, but the first call to
    any of its methods will concretize that choice - either as
    the default, or the backend set via `TaskManager.set_backend`.
    It cannot be changed after this point.
    """

    _backend: Backend | None = None

    def set_backend(self, backend: Backend) -> Backend:
        """Choose a backend.

        Sets the backend, provided it has not already been loaded.

        Args:
            backend: chosen backend, to override the default.

        Returns:
            Backend that was set.

        Raises:
            RuntimeError: when a backend has already been loaded.
        """
        if self._backend is not None:
            raise RuntimeError(
                f"Backend is already loaded ({self._backend}). Did an imported module use lazy/run already?"
            )
        self._backend = backend
        return self._backend

    @cached_property
    def backend(self) -> BackendModule:
        """Import backend module.

        Cached property to load the backend module, if it has not been already.

        Returns:
            Backend module for the specific choice of backend.
        """
        backend = self._backend
        if backend is None:
            backend = self.set_backend(DEFAULT_BACKEND)

        backend_mod = importlib.import_module(
            f".backends.backend_{backend.value}", "dewret"
        )
        return backend_mod

    def make_lazy(self) -> LazyFactory:
        """Get the lazy decorator for this backend.

        Returns:
            Real decorator for this backend.
        """
        return self.backend.lazy

    def evaluate(
        self,
        task: Lazy | list[Lazy] | tuple[Lazy],
        __workflow__: Workflow,
        thread_pool: ThreadPoolExecutor | None = None,
        **kwargs: Any,
    ) -> Any:
        """Evaluate a single task for a known workflow.

        Args:
            task: the task to evaluate.
            __workflow__: workflow within which this exists.
            thread_pool: existing pool of threads to run this in, or None.
            **kwargs: any arguments to pass to the task.
        """
        result = self.backend.run(__workflow__, task, thread_pool=thread_pool, **kwargs)
        new_result, collected_workflow = unify_workflows(result, __workflow__)

        if collected_workflow is None:
            raise RuntimeError("A new workflow could not be found")

        # Then we set the result to be the whole thing
        collected_workflow.set_result(new_result)
        return collected_workflow.result

    def unwrap(self, task: Lazy) -> Target:
        """Unwraps a lazy-evaluated function to get the function.

        Ideally, we could use the `__wrapped__` property but not all
        workflow engines support this, and most importantly, dask has
        only done so as of 2024.03.

        Args:
            task: task to be unwrapped.

        Returns:
            Original target.

        Raises:
            RuntimeError: if the task is not a wrapped function.
        """
        return self.backend.unwrap(task)

    def ensure_lazy(self, task: Any) -> Lazy | None:
        """Evaluate a single task for a known workflow.

        As we mask our lazy-evaluable functions to appear as their original
        types to the type system (see `dewret.tasks.task`), we must cast them
        back, to allow the type-checker to comb the remainder of the code.

        Args:
            task: the suspected task to check.

        Returns:
            Original task, cast to a Lazy, or None.
        """
        if isinstance(task, LazyEvaluation):
            return self.ensure_lazy(task._fn)
        return task if self.backend.is_lazy(task) else None

    def __call__(
        self,
        task: Any,
        __workflow__: Workflow | None = None,
        **kwargs: Unpack[ConstructConfigurationTypedDict],
    ) -> Workflow:
        """Execute the lazy evalution.

        Arguments:
            task: the task to evaluate.
            simplify_ids: when we finish running, make nicer step names?
            **kwargs: any arguments to pass to the task.

        Returns:
            A reusable reference to this individual step.
        """
        workflow = __workflow__ or Workflow()

        with set_configuration(**kwargs):
            context = copy_context().items()

            def _initializer() -> None:
                for var, value in context:
                    var.set(value)

            thread_pool = ThreadPoolExecutor(initializer=_initializer)

            result = self.evaluate(task, workflow, thread_pool=thread_pool, **kwargs)
            simplify_ids = bool(get_configuration("simplify_ids"))
        return Workflow.from_result(result, simplify_ids=simplify_ids)


_manager = TaskManager()
lazy = _manager.make_lazy
ensure_lazy = _manager.ensure_lazy
unwrap = _manager.unwrap
evaluate = _manager.evaluate
construct = _manager
"""An alias pointing to an instance of the TaskManager class.
Used for constructing a set of tasks into a dewret workflow instance.
"""


class TaskException(Exception):
    """Exception tied to a specific task.

    Primarily aimed at parsing issues, but this will ensure that
    a message is shown with useful debug information for the
    workflow writer.
    """

    def __init__(
        self,
        task: Task | Target,
        dec_tb: TracebackType | None,
        tb: TracebackType | None,
        message: str,
        *args: Any,
        **kwargs: Any,
    ):
        """Create a TaskException for this exception.

        Args:
            task: the Task causing the exception.
            dec_tb: a traceback of the task declaration.
            tb: a traceback of the original task call.
            message: a message to show to the user.
            *args: any other arguments accepted by Exception.
            **kwargs: any other arguments accepted by Exception.
        """
        if dec_tb:
            frame = traceback.extract_tb(dec_tb)[-1]
            self.add_note(
                f"Task {task.__name__} declared in {frame.name} at {frame.filename}:{frame.lineno}\n"
                f"{frame.line}"
            )
        super().__init__(message)
        self.__traceback__ = tb


_IN_NESTED_TASK: ContextVar[bool] = ContextVar("in_nested_task")
_IN_NESTED_TASK.set(False)


@contextmanager
def in_nested_task() -> Generator[None, None, None]:
    """Informs the builder that we are within a nested task.

    This is only really relevant in the subworkflow context.

    TODO: check impact of ContextVar being thread-sensitive on build.
    """
    try:
        tok = _IN_NESTED_TASK.set(True)
        yield
    finally:
        _IN_NESTED_TASK.reset(tok)


def is_in_nested_task() -> bool:
    """Check if we are within a nested task.

    Used, for example, to see if discovered parameters should be
    treated as "local" (i.e. should take a default to the step) or
    global (i.e. should be turned into a workflow parameter) if we
    are inside or outside a subworkflow, respectively.
    """
    try:
        return _IN_NESTED_TASK.get()
    except LookupError:
        return False


def factory(fn: Callable[..., RetType]) -> Callable[..., RetType]:
    """Create a factory, that can be treated as complex input to a workflow.

    Args:
        fn: a callable to create the entity.
    """
    return task(is_factory=True)(fn)


# Workaround for PyCharm
factory: Callable[[Callable[..., RetType]], Callable[..., RetType]] = factory


def workflow() -> Callable[[Callable[Param, RetType]], Callable[Param, RetType]]:
    """Shortcut for marking a task as nested.

    A nested task is one which calls other tasks and does not
    do anything else important. It will _not_ actually get called
    at runtime, but should map entirely into the graph. As such,
    arithmetic operations on results, etc. will cause errors at
    render-time. Combining tasks is acceptable, and intended. The
    effect of the nested task will be considered equivalent to whatever
    reaching whatever step reference is returned at the end.

    ```python
    >>> @task()
    ... def increment(num: int) -> int:
    ...     return num + 1

    >>> @workflow()
    ... def double_increment(num: int) -> int:
    ...     return increment(increment(num=num))

    ```

    Returns:
        Task that runs at render, not execution, time.
    """
    return task(nested=True, flatten_nested=False)


# Workaround for PyCharm
workflow: Callable[[], Callable[[T], T]] = workflow


def task(
    nested: bool = False,
    flatten_nested: bool = True,
    is_factory: bool = False,
) -> Callable[[Callable[Param, RetType]], Callable[Param, RetType]]:
    """Decorator factory abstracting backend's own task decorator.

    For example:

    ```python
    >>> @task()
    ... def increment(num: int) -> int:
    ...     return num + 1

    ```

    If the backend is `dask` (the default), it is will evaluate this
    as a `dask.delayed`. Note that, with any backend, dewret will
    hijack the decorator to record the attempted _evalution_ rather than
    actually _evaluating_ the lazy function. Nonetheless, this hijacking
    will still be executed with the backend's lazy executor, so
    `dask.delayed` will still be called, for example, in the dask case.

    Args:
        nested: whether this should be executed to find other tasks.
        flatten_nested: (only relevant to nested tasks) should this nested task
            be considered a distinct subworkflow, or is it just organizational
            for the outer workflow.
        is_factory: whether this task should be marked as a 'factory', rather than
            a normal step.

    Returns:
        Decorator for the current backend to mark lazy-executable tasks.

    Raises:
        TypeError: if arguments are missing or incorrect, in line with usual
            Python behaviour.
    """

    def _task(fn: Callable[Param, RetType]) -> Callable[Param, RetType]:
        declaration_tb = make_traceback()

        def _fn(
            *args: Any,
            __workflow__: Workflow | None = None,
            __traceback__: TracebackType | None = None,
            **kwargs: Any,
        ) -> RetType:
            configuration = None
            allow_positional_args = bool(get_configuration("allow_positional_args"))

            try:
                # Ensure that all arguments are passed as keyword args and prevent positional args.
                # passed at all.
                if args and not allow_positional_args:
                    raise TypeError(
                        f"""
                        Calling {fn.__name__}: Arguments must _always_ be named,
                        e.g. my_task(num=1) not my_task(1)\n"

                        @task()
                        def add_numbers(left: int, right: int):
                            return left + right

                        construct(add_numbers(left=3, right=5))
                        """
                    )

                # Ensure that the passed arguments are, at least, a Python-match for the signature.
                sig = inspect.signature(fn)
                positional_args = {key: False for key in kwargs}
                for arg, (key, _) in zip(args, sig.parameters.items(), strict=False):
                    if isinstance(arg, IteratedGenerator):
                        for inner_arg, (key, _) in zip(
                            arg, sig.parameters.items(), strict=False
                        ):
                            if key in positional_args:
                                continue
                            kwargs[key] = inner_arg
                            positional_args[key] = True
                    else:
                        kwargs[key] = arg
                        positional_args[key] = True
                sig.bind(**kwargs)

                def _to_param_ref(value: Any) -> ParameterReference[Any] | None:
                    if isinstance(value, Parameter):
                        return value.make_reference(workflow=__workflow__)
                    return None

                refs = []
                for key, val in kwargs.items():
                    val, kw_refs = expr_to_references(val, remap=_to_param_ref)
                    refs += kw_refs
                    kwargs[key] = val
                # Not realistically going to be other than Workflow.
                workflows: list[Workflow] = [
                    cast(Workflow, reference.__workflow__)
                    for reference in refs
                    if hasattr(reference, "__workflow__")
                    and reference.__workflow__ is not None
                ]
                if __workflow__ is not None:
                    workflows.insert(0, __workflow__)
                if workflows:
                    workflow = Workflow.assimilate(*workflows)
                else:
                    workflow = Workflow()

                analyser = FunctionAnalyser(fn)

                if not is_in_nested_task():
                    for var, value in kwargs.items():
                        if analyser.is_at_construct_arg(var):
                            kwargs[var] = value
                        elif is_firm(value):
                            # We leave this reference dangling for a consumer to pick up ("tethered"), unless
                            # we are in a nested task, that does not have any existence of its own.
                            tethered: Literal[False] | None = (
                                False
                                if nested
                                and (
                                    flatten_nested
                                    or get_configuration("flatten_all_nested")
                                )
                                else None
                            )
                            kwargs[var] = cast(
                                Parameter[Any],
                                param(
                                    var,
                                    value,
                                    tethered=tethered,
                                    autoname=tethered is not False,
                                    typ=analyser.get_argument_annotation(var) or UNSET,
                                ),
                            ).make_reference(workflow=workflow)
                original_kwargs = dict(kwargs)
                fn_globals = analyser.globals

                for var, value in fn_globals.items():
                    # This error is redundant as it triggers a SyntaxError in Python.
                    # Note: the following test duplicates a syntax error.
                    #   if var in kwargs:
                    #       raise TypeError(
                    #           "Captured parameter {var} (global variable in task) shadows an argument"
                    #        )
                    if isinstance(value, Parameter):
                        kwargs[var] = value.make_reference(workflow=workflow)
                    elif is_task(value) or ensure_lazy(value) is not None:
                        if not nested and _workaround_check_value_is_task(
                            fn, var, value
                        ):
                            raise TypeError(
                                f"""
                                You referenced a task {var} inside another task {fn.__name__}, but it is not a workflow
                                - this will not be found!

                                @task()
                                def {var}(...) -> ...:
                                    ...

                                @workflow() <<<--- likely what you want
                                def {fn.__name__}(...) -> ...:
                                    ...
                                    {var}(...)
                                    ...
                                """
                            )
                    # If nested, we will execute the insides, and it is reasonable and important
                    # to have a full set of annotations for any encountered variables.
                    elif (
                        nested
                        and not analyser.get_argument_annotation(var, exhaustive=True)
                        and not inspect.isclass(value)
                        or inspect.isfunction(value)
                    ):
                        raise RuntimeError(
                            f"Could not find a type annotation for {var} for {fn.__name__}"
                        )
                    elif (
                        analyser.is_at_construct_arg(var, exhaustive=True)
                        or value is evaluate
                        or value is construct  # Allow manual building.
                    ):
                        kwargs[var] = value
                    elif inspect.isclass(value) or inspect.isfunction(value):
                        # We assume these are loaded at runtime.
                        ...
                    elif is_firm(value) or (
                        (attrs_has(value) or is_dataclass(value))
                        and not inspect.isclass(value)
                    ):
                        kwargs[var] = cast(
                            Parameter[Any],
                            param(
                                var,
                                default=value,
                                tethered=False,
                                typ=analyser.get_argument_annotation(
                                    var, exhaustive=True
                                )
                                or UNSET,
                            ),
                        ).make_reference(workflow=workflow)
                    elif (
                        is_expr(value)
                        and (expr_refs := expr_to_references(value))
                        and len(expr_refs[1]) != 0
                    ):
                        kwargs[var] = value
                    elif nested:
                        raise NotImplementedError(
                            f"Nested tasks must now only refer to global parameters, raw or tasks, not objects: {var}"
                        )
                if nested:
                    if flatten_nested or get_configuration("flatten_all_nested"):
                        with in_nested_task():
                            output = analyser.with_new_globals(kwargs)(
                                **original_kwargs
                            )
                        lazy_fn = ensure_lazy(output)
                        if lazy_fn is not None:
                            with in_nested_task():
                                output = evaluate(lazy_fn, __workflow__=workflow)
                            # raise TypeError(
                            #    f"Task {fn.__name__} returned output of type {type(output)}, which is not a lazy function for this backend."
                            # )
                        step_reference = output
                    else:
                        nested_workflow = Workflow(name=fn.__name__)
                        nested_globals: dict[str, Any] = {
                            var: cast(
                                Parameter[Any],
                                param(
                                    var,
                                    default=value.__default__
                                    if hasattr(value, "__default__")
                                    else UNSET,
                                    typ=(value.__type__),
                                    tethered=nested_workflow,
                                ),
                            ).make_reference(workflow=nested_workflow)
                            if isinstance(value, Reference)
                            else value
                            for var, value in kwargs.items()
                        }
                        nested_kwargs = {
                            key: value
                            for key, value in nested_globals.items()
                            if key in original_kwargs
                        }
                        with in_nested_task():
                            output = analyser.with_new_globals(nested_globals)(
                                **nested_kwargs
                            )
                            nested_workflow = _manager(
                                output, __workflow__=nested_workflow
                            )
                        step_reference = workflow.add_nested_step(
                            fn.__name__,
                            nested_workflow,
                            analyser.return_type,
                            original_kwargs,
                            positional_args,
                        )
                    if is_expr(step_reference):
                        return cast(RetType, step_reference)
                    raise TypeError(
                        f"Nested tasks must return a step reference, not {type(step_reference)} to ensure graph makes sense."
                    )
                step = cast(
                    RetType,
                    workflow.add_step(
                        fn,
                        kwargs,
                        raw_as_parameter=not is_in_nested_task(),
                        is_factory=is_factory,
                        positional_args=positional_args,
                    ),
                )
                return step
            except TaskException as exc:
                raise exc
            except Exception as exc:
                raise TaskException(
                    fn,
                    declaration_tb,
                    __traceback__,
                    exc.args[0] if exc.args else f"Could not call task {fn.__name__}",
                ) from exc
            finally:
                if configuration:
                    configuration.__exit__(None, None, None)

        _fn.__step_expression__ = True  # type: ignore
        _fn.__original__ = fn  # type: ignore
        return LazyEvaluation(_fn)

    return _task


# Workaround for PyCharm
task: Callable[[], Callable[[T], T]] = task


def set_backend(backend: Backend) -> None:
    """Choose a backend.

    Will raise an error if a backend is already chosen.

    Args:
        backend: chosen backend to use from here-on in.
    """
    _manager.set_backend(backend)


def _workaround_check_value_is_task(
    fn: Callable[..., Any], var: str, value: Any
) -> bool:
    """Workaround for known issue in Python stdlib.

    getclosurevars does not distinguish between globals and inline-imported
    variables. For task-based workflows, this is a much more import use-case
    than in regular Python code, and the bug has been open since 2019, so a
    workaround seems the best course of action.

    This attempts to limit the false positives, by ensuring the "global"
    is not a task in an already-imported module (at least not
    if imported the same name). See `task_errors.py` for more limitations.

    See: https://bugs.python.org/issue36697
    """
    return all(
        (var not in vars(sys.modules[mod]) or is_task(vars(sys.modules[mod])[var]))
        for mod in list(fn.__code__.co_names) + list(fn.__code__.co_varnames)
        if mod in sys.modules
    )



from dewret.tasks import task, workflow

from .other import nothing

JUMP: float = 1.0
test: float = nothing


@workflow()
def try_nothing() -> int:
    """Check that we can see AtRender in another module."""
    if nothing:
        return increment(num=1)
    return increment(num=0)


@task()
def increase(num: int | float) -> float:
    """Add 1 to a number."""
    return num + JUMP


@task()
def increment(num: int) -> int:
    """Increment an integer."""
    return num + 1


@task()
def double(num: int | float) -> int | float:
    """Double an integer."""
    return 2 * num


@task()
def mod10(num: int) -> int:
    """Remainder of an integer divided by 10."""
    return num % 10


@task()
def sum(left: int | float, right: int | float) -> int | float:
    """Add two integers."""
    return left + right


@task()
def pi() -> float:
    """Returns pi."""
    import math

    return math.pi


@workflow()
def triple_and_one(num: int | float) -> int | float:
    """Triple a number by doubling and adding again, then add 1."""
    return sum(left=sum(left=double(num=num), right=num), right=1)


@task()
def tuple_float_return() -> tuple[float, float]:
    """Return a tuple of floats."""
    return 48.856667, 2.351667


@task()
def reverse_list(to_sort: list[int | float]) -> list[int | float]:
    return to_sort[::-1]


@task()
def max_list(lst: list[int | float]) -> int | float:
    return max(lst)



# Copyright 2014 Flax & Teal Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Snakemake Renderer.

Outputs a [Snakemake](https://snakemake.github.io/) representation of the
current workflow.

"""

import os
import yaml
import inspect
import typing

from attrs import define

from dewret.core import (
    Raw,
    BasicType,
    Reference,
)
from dewret.workflow import (
    Workflow,
    Task,
    Lazy,
    BaseStep,
)
from dewret.render import (
    base_render,
)

MainTypes = BasicType | list[str] | list["MainTypes"] | dict[str, "MainTypes"]


@define
class ReferenceDefinition:
    """Represents a Snakemake-renderable internal reference.

    Attributes:
        source (str): The source of the internal reference.

    Methods:
        from_reference(cls, ref: Reference) -> "ReferenceDefinition": Constructs a
            ReferenceDefinition object from a Reference object, extracting the source
            of the reference.
        render(self) -> str: Renders the internal reference definition as a string
            suitable for use in Snakemake workflows.
    """

    source: str

    @classmethod
    def from_reference(cls, ref: Reference[typing.Any]) -> "ReferenceDefinition":
        """Build from a `Reference`.

        Converts a `dewret.workflow.Reference` into a Snakemake-rendering object.

        Args:
            ref: reference to convert.
        """
        return cls(source=ref.name)

    def render(self) -> str:
        """Render the internal reference definition as a string.

        Returns:
            str: internal reference.
        """
        return self.source


# TODO: Refactor: better way to handle the types.
def to_snakemake_type(param: Raw) -> str:
    """Convert a raw type to a corresponding Snakemake-compatible Python type.

    This function maps raw types to their corresponding Python types as
    used in Snakemake. Snakemake primarily uses Python types for its parameters,
    and this function ensures that the provided type is appropriately converted.

    Args:
        param: The raw type to be converted, which can be of any type.

    Returns:
        A string representing the corresponding Python type for Snakemake.

    Raises:
        TypeError: If the parameter's type cannot be mapped to a known Python type.
    """
    typ = str(param)
    if typ.__contains__("str"):
        return f'"{typ.replace("str|", "")}"'
    elif typ.__contains__("bool"):
        return typ.replace("bool|", "")
    elif typ.__contains__("dict"):
        return typ.replace("dict|", "")
    elif typ.__contains__("list"):
        return typ.replace("list|", "")
    elif typ.__contains__("float"):
        return typ.replace("float|", "")
    elif typ.__contains__("int"):
        return typ.replace("int|", "")
    else:
        raise TypeError(f"Cannot render complex type ({typ})")


def get_method_args(func: Lazy) -> inspect.Signature:
    """Retrieve the argument names and types of a lazy-evaluatable function.

    Args:
        func: A function that adheres to the `Lazy` protocol.

    Returns:
        An `ItemsView` object containing the argument names and their corresponding
        `inspect.Parameter` objects.
    """
    args = inspect.signature(func)

    return args


def get_method_rel_path(func: Lazy) -> typing.Any:
    """Get the relative path of the module containing the given function.

    Args:
        func (Lazy): The function for which the relative path is to be determined.

    Returns:
        any: The relative path of the module containing the function.

    Note:
        This function relies on the `inspect` and `os` modules to compute its relative path.
    """
    # TODO: error handling
    source_file = inspect.getsourcefile(func)
    if source_file:
        relative_path = os.path.relpath(source_file, start=os.getcwd())
        module_name = os.path.splitext(relative_path)[0].replace(os.path.sep, ".")

    return module_name


@define
class InputDefinition:
    """Represents input and parameter definitions block for a Snakemake-renderable workflow step.

    Attributes:
        inputs (List[str]): A list of input definitions.
        params (List[str]): A list of parameter definitions.

    Methods:
        from_step(cls, step: Step) -> "InputDefinition": Constructs an InputDefinition
            object from a Step object, extracting inputs and parameters and converting
            them to Snakemake-compatible format.
        render(self) -> dict[str, str]: Renders the input and parameter definitions
            as a dictionary for use in Snakemake Input and Params blocks.
    """

    inputs: list[str]
    params: list[str]

    @classmethod
    def from_step(cls, step: BaseStep) -> "InputDefinition":
        """Constructs an InputDefinition object from a Step.

        Args:
            step (Step): The Step object from which input and parameter block definitions are
                extracted.

        Returns:
            InputDefinition: An InputDefinition object.
        """
        params = []
        inputs = []
        for key, param in step.arguments.items():
            if isinstance(param, Reference):
                ref = (
                    ReferenceDefinition.from_reference(param)
                    .render()
                    .replace("-", "_")
                    .replace("/out", ".output")
                )
                input = f"{key}=rules.{ref}.output_file"
                inputs.append(input)
                params.append(input + ",")
            elif isinstance(param, Raw):
                customized = f"{key}={to_snakemake_type(param)},"
                params.append(customized)

        if params:
            params[len(params) - 1] = params[len(params) - 1].replace(",", "")

        return cls(inputs=inputs, params=params)

    def render(self) -> dict[str, list[str]]:
        """Renders the input and parameter definitions as a dictionary.

        Returns:
            dict[str, list[MainTypes]]: A dictionary containing the input and parameter definitions,
                for use in Snakemake Input and Params blocks.
        """
        return {"inputs": self.inputs, "params": self.params}


@define
class OutputDefinition:
    """Represents the output definition block for a Snakemake-renderable workflow step.

    Attributes:
        output_file (str): The output file definition.

    Methods:
        from_step(cls, step: Step) -> "OutputDefinition": Constructs an OutputDefinition
            object from a Step object, extracting and converting the output file definition
            to Snakemake-compatible format.

        render(self) -> list[str]: Renders the output definition as a list
            suitable for use in Snakemake Output block.
    """

    output_file: str

    @classmethod
    def from_step(cls, step: BaseStep) -> "OutputDefinition":
        """Constructs an OutputDefinition object from a Step.

        Args:
            step (Step): The Step object from which the output file definition is extracted.

        Returns:
            OutputDefinition: An OutputDefinition object, for use in Snakemake Output block.
        """
        # TODO: Error handling
        # TODO: Better way to handling input/output files
        output_file = step.arguments.get("output_file", Raw("OUTPUT_FILE"))
        if isinstance(output_file, Raw):
            args = to_snakemake_type(output_file)

        return cls(output_file=args)

    def render(self) -> list[str]:
        """Renders the output definition as a list.

        Returns:
            list[str]: A list containing the output file definition, for use in a Snakemake Output block.
        """
        # The comma after the last element is mandatory for the structure of rule onces it's used in yaml.dump
        return [
            f"output_file={self.output_file}",
        ]


@define
class RunDefinition:
    """Represents a Snakemake-renderable run block for a dewret workflow step.

    Attributes:
        method_name (str): The name of the method to be executed in the snakefile run block.
        rel_import (str): The relative import path of the method.
        args (List[str]): The arguments to be passed to the method.

    Methods:
        from_task(cls, task: Task) -> "RunDefinition": Constructs a RunDefinition
            object from a Task object, extracting method information and arguments
            from the task and converting them to Snakemake-compatible format.

        render(self) -> list[str]: A list containing the import statement and the method
            call statement, for use in Snakemake run block.
    """

    method_name: str
    rel_import: str
    args: list[str]

    @classmethod
    def from_task(cls, task: Task | Workflow) -> "RunDefinition":
        """Constructs a RunDefinition object from a Task.

        Args:
            task (Task): The Task object from which method information and arguments
                are extracted.

        Returns:
            RunDefinition: A RunDefinition object containing the converted method
                information and arguments.
        """
        if isinstance(task, Task):
            relative_path = get_method_rel_path(task.target)
            rel_import = f"{relative_path}"
            args = get_method_args(task.target)
        else:
            # TODO: Add implementation for multiple workflows
            pass

        signature = [
            f"{param_name}=params.{param_name}" for param_name in args.parameters.keys()
        ]

        return cls(method_name=task.name, rel_import=rel_import, args=signature)

    def render(self) -> list[str]:
        """Renders the run block as a list of strings.

        Returns:
            list[str]: A list containing the import statement and the method
                call statement, for use in Snakemake run block.
        """
        signature = ", ".join(f"{arg}" for arg in self.args)
        # The comma after the last element is mandatory for the structure of rule onces it's used in yaml.dump
        return [
            f"import {self.rel_import}\n",
            f"{self.rel_import}.{self.method_name}({signature})\n",
        ]


@define
class StepDefinition:
    """Represents a Snakemake-renderable step definition in a dewret workflow.

    Attributes:
        name (str): The name of the step.
        run (str): The run block definition for the step.
        params (List[str]): The parameter definitions for the step.
        output (list[str]: The output definition for the step.
        input (List[str]): The input definitions for the step.

    Methods:
        from_step(cls, step: Step) -> "StepDefinition": Constructs a StepDefinition
            object from a Step object, extracting step information and components
            from the step and converting them to Snakemake format.
        render(self) -> dict[str, MainTypes]: Renders the step definition as a dictionary
            suitable for use in Snakemake workflows.
    """

    name: str
    run: list[str]
    params: list[str]
    output: list[str]
    input: list[str]

    @classmethod
    def from_step(cls, step: BaseStep) -> "StepDefinition":
        """Constructs a StepDefinition object from a Step.

        Args:
            step (Step): The Step object from which step information and components
                are extracted.

        Returns:
            StepDefinition: A StepDefinition object containing the converted step
                information and components.
        """
        input_block = InputDefinition.from_step(step).render()
        run_block = RunDefinition.from_task(step.task).render()
        output_block = OutputDefinition.from_step(step).render()
        return cls(
            name=step.name,
            run=run_block,
            params=input_block["params"],
            input=input_block["inputs"],
            output=output_block,
        )

    def render(self) -> dict[str, MainTypes]:
        """Renders the step definition as a dictionary.

        Returns:
            dict[str, MainTypes]: A dictionary containing the components of the step
                definition, for use in Snakemake workflows.
        """
        return {
            "run": self.run,
            "input": self.input,
            "params": self.params,
            "output": self.output,
        }


# TODO: Find out why the yaml.dump scrambles the order of the rules
# TODO: Add a rule all: with input definition the last outputed file


@define
class WorkflowDefinition:
    """Represents a Snakemake-renderable workflow definition from a dewret workflow.

    Attributes:
        steps (List[StepDefinition]): A list of StepDefinition objects representing the steps
            in the workflow.

    Methods:
        from_workflow(cls, workflow: Workflow) -> "WorkflowDefinition": Constructs a WorkflowDefinition
            object from a Workflow object, converting its steps to StepDefinition objects.

        render(self) -> dict[str, MainTypes]: Renders the workflow definition as a dictionary
            containing Snakemake rules, for use in Snakemake workflows.
    """

    steps: list[StepDefinition]

    # TODO: Add a "rule all" definition
    # In order for snakemake to execute multiple connected rules
    # it needs a rule to define which is the target rule that connects all the multiple rules
    @classmethod
    def from_workflow(cls, workflow: Workflow) -> "WorkflowDefinition":
        """Creates a WorkflowDefinition object from a Workflow object.

        Args:
            workflow (Workflow): The workflow to be rendered.

        Returns:
            str: WorkflowDefinition with definited steps.
        """
        return cls(steps=[StepDefinition.from_step(step) for step in workflow.steps])

    def render(self) -> dict[str, MainTypes]:
        """Render the WorkflowDefinition.

        Returns:
            dict[str, MainTypes]: A dictionary containing the components of the Workflow
                definition, for use in Snakemake workflows.
        """
        return {
            f"rule {step.name.replace('-', '_')}": step.render() for step in self.steps
        }


def raw_render(workflow: Workflow) -> dict[str, MainTypes]:
    """Render the workflow as a Snakemake (SMK) string.

    This function converts a Workflow object into a object containing snakemake rules.

    Args:
        workflow (Workflow): The workflow to be rendered.

    Returns:
        dict[str, MainTypes]: A dictionary containing the components of the Workflow
            definition, for use in Snakemake workflows.
    """
    return WorkflowDefinition.from_workflow(workflow).render()


def render(workflow: Workflow) -> dict[str, typing.Any]:
    """Render the workflow as a Snakemake (SMK) string.

    This function converts a Workflow object into a Snakemake-compatible yaml.

    Args:
        workflow (Workflow): The workflow to be rendered.

    Returns:
        str: A Snakemake-compatible yaml representation of the workflow.
    """
    trans_table = str.maketrans(
        {
            "-": "   ",
            "'": "",
            "[": "",
            "]": "",
        }
    )

    return base_render(
        workflow,
        lambda workflow: yaml.dump(
            WorkflowDefinition.from_workflow(workflow).render(), indent=4
        ).translate(trans_table),
    )



"""Basic snakemake workflow.

Useful as an example of a workflow with multiple tasks.

```sh
$ python snakemake_workflow.py
```
Since snakemake commonly communicates via the filesystem between workflow steps,
each our tasks and hence steps writes its output to a file.
This output can then be read by subsequent steps.

For the current implementation, output files must be called "output_file".

Running the example will generate a Snakefile with the following code:

rule download_data_cbb4d1260b8b17d6693986ebfd833d53:
    input:
    output:
        output_file="generated_smk/data/test_data.txt"
    params:
        data="4, 5, 7, 8, 0",
        output_file="generated_smk/data/test_data.txt"
    run:
        import snakemake_tasks


        snakemake_tasks.download_data(data=params.data, output_file=params.output_file)


rule generate_report_b1f9747ed9903dcd9571246547b1767b:
    input:
        processed_data=rules.process_data_75213b5c31fed773cd1f7ca1dfb6ab5a.output.output_file
    output:
        output_file="results/report.txt"
    params:
        processed_data=rules.process_data_75213b5c31fed773cd1f7ca1dfb6ab5a.output.output_file,
        multiple_arg="test2",
        output_file="results/report.txt"
    run:
        import snakemake_tasks


        snakemake_tasks.generate_report(processed_data=params.processed_data, multiple_arg=params.multiple_arg,
        output_file=params.output_file)


rule process_data_75213b5c31fed773cd1f7ca1dfb6ab5a:
    input:
        data_file=rules.download_data_cbb4d1260b8b17d6693986ebfd833d53.output.output_file
    output:
        output_file="generated_smk/data/processed_data.txt"
    params:
        data_file=rules.download_data_cbb4d1260b8b17d6693986ebfd833d53.output.output_file,
        multiple_arg="test1",
        output_file="generated_smk/data/processed_data.txt"
    run:
        import snakemake_tasks


        snakemake_tasks.process_data(data_file=params.data_file, multiple_arg=params.multiple_arg,
        output_file=params.output_file)

In order to make the snakemake file be executable you need to:
- manually add:
rule all:
    input:
        "results/report.txt" # To make sure all connected rules are executed. For more information https://snakemake.readthedocs.io/
- Change the order of the rules to match the order of execution.
- Make sure all methods in run blocks are on the same line.
- Remove @task annotation from the snakemake_workflow.py
"""

import yaml
from dewret.tasks import task, construct
from dewret.renderers.snakemake import render


@task()
def download_data(data: str, output_file: str) -> str:
    """Example task to be converted to snakemake workflow rule.

    This function writes the given data to the specified output file.

    Args:
        data (str): The data to be written to the output file.
        output_file (str): The path to the output file where the data will be written.

    Returns:
        str: The path to the output file.
    """
    with open(output_file, "w") as f:
        f.write(data)

    return output_file


@task()
def process_data(data_file: str, multiple_arg: str, output_file: str) -> str:
    """Example task to be converted to snakemake workflow rule.

    This function reads data from a file, processes it by appending additional
    strings, and writes the result to an output file.

    Args:
        data_file (str): The path to the input file containing the data to be processed.
        multiple_arg (str): An additional string to be appended during processing.
        output_file (str): The path to the output file where the processed data will be written.

    Returns:
        str: The path to the output file.
    """
    new_data = "test"

    with open(data_file, "r") as f:
        data = f.read()
        new_data = data + new_data + multiple_arg

    with open(output_file, "w") as f:
        f.write(new_data)

    return output_file


@task()
def generate_report(processed_data: str, multiple_arg: str, output_file: str) -> str:
    """Example task to be converted to snakemake workflow.

    This function generates a report by reading processed data from a file and
    appending an additional string, then writes the report to an output file.

    Args:
        processed_data (str): The path to the input file containing the processed data.
        multiple_arg (str): An additional string to be appended to the report.
        output_file (str): The path to the output file where the report will be written.

    Returns:
        str: The path to the output file.
    """
    with open(processed_data, "r") as datafile:
        data = datafile.read()

    with open(output_file, "w") as f:
        f.write(data + multiple_arg)

    return output_file


if __name__ == "__main__":
    data = download_data(
        data="4, 5, 7, 8, 0", output_file="generated_smk/data/test_data.txt"
    )
    processed_data = process_data(
        data_file=data,
        multiple_arg="test1",
        output_file="generated_smk/data/processed_data.txt",
    )
    result = generate_report(
        processed_data=processed_data,
        multiple_arg="test2",
        output_file="results/report.txt",
    )

    workflow = construct(result)
    smk_output = render(workflow)

    with open("Snakefile", "w") as file:
        trans_table = str.maketrans(
            {
                "-": "   ",
                "'": "",
                "[": "",
                "]": "",
            }
        )
        smk_text = yaml.dump(smk_output, indent=4).translate(trans_table)
        file.write(smk_text)



"""Test for expected errors."""

import pytest
from dewret.workflow import Task, Lazy
from dewret.tasks import construct, task, workflow, TaskException
from dewret.annotations import AtRender
from ._lib.extra import increment, pi, reverse_list  # noqa: F401


@task()  # This is expected to be the line number shown below.
def add_task(left: int, right: int) -> int:
    """Adds two values and returns the result."""
    return left + right


ADD_TASK_LINE_NO: int = 10


@workflow()
def badly_add_task(left: int, right: int) -> int:
    """Badly attempts to add two numbers."""
    return add_task(left=left)  # type: ignore


@task()
def badly_wrap_task() -> int:
    """Sums two values but should not be calling a task."""
    return add_task(left=3, right=4)


class MyStrangeClass:
    """Dummy class for tests."""

    def __init__(self, task: Task):
        """Dummy constructor for tests."""
        ...


@task()
def pi_exported_from_math() -> float:
    """Get pi from math package by name."""
    from math import pi

    return pi


@task()
def try_recursive() -> float:
    """Get pi from math package by name."""
    return try_recursive()


@task()
def pi_hidden_by_math() -> float:
    """Recursive call with a bug, that we cannot spot because the math library confounds our check."""
    import math  # noqa: F401

    return pi()


@task()
def pi_hidden_by_math_2() -> float:
    """Recursive call with a bug, that we cannot spot because the math library confounds our check."""
    math = 1  # noqa: F841
    return pi()  # noqa: F821


@task()
def pi_with_visible_module_task() -> float:
    """Imported task that _will_ be spotted."""
    from ._lib import extra

    return extra.increment(2)


@task()
def pi_with_invisible_module_task() -> float:
    """Imported task that will _not_ be spotted."""
    from ._lib import extra

    return extra.double(3.14 / 2)


@workflow()
def unacceptable_object_usage() -> int:
    """Invalid use of custom object within nested task."""
    return MyStrangeClass(add_task(left=3, right=4))  # type: ignore


@workflow()
def unacceptable_nested_return(int_not_global: AtRender[bool]) -> int | Lazy:
    """Bad subworkflow that fails to return a task."""
    add_task(left=3, right=4)
    return 7 if int_not_global else ADD_TASK_LINE_NO


def test_missing_arguments_throw_error() -> None:
    """Check whether omitting a required argument will give an error.

    Since we do not run the original function, it is up to dewret to check
    that the signature is, at least, acceptable to Python.

    WARNING: in keeping with Python principles, this does not error if types
    mismatch, but `mypy` should. You **must** type-check your code to catch these.
    """
    with pytest.raises(TaskException) as exc:
        add_task(left=3)  # type: ignore
    end_section = str(exc.getrepr())[-500:]
    assert str(exc.value) == "missing a required argument: 'right'"
    assert "Task add_task declared in <module> at " in end_section
    assert f"test_errors.py:{ADD_TASK_LINE_NO}" in end_section


def test_missing_arguments_throw_error_in_subworkflow() -> None:
    """Check whether omitting a required argument within a subworkflow will give an error.

    Since we do not run the original function, it is up to dewret to check
    that the signature is, at least, acceptable to Python.

    WARNING: in keeping with Python principles, this does not error if types
    mismatch, but `mypy` should. You **must** type-check your code to catch these.
    """
    with pytest.raises(TaskException) as exc:
        badly_add_task(left=3, right=4)
    end_section = str(exc.getrepr())[-500:]
    assert str(exc.value) == "missing a required argument: 'right'"
    assert "def badly_add_task" in end_section
    assert "Task add_task declared in <module> at " in end_section
    assert f"test_errors.py:{ADD_TASK_LINE_NO}" in end_section


def test_positional_arguments_throw_error() -> None:
    """Check whether unnamed (positional) arguments throw an error.

    We can use default and non-default arguments, but we expect them
    to _always_ be named.
    """
    with pytest.raises(TaskException) as exc:
        add_task(3, right=4)
    assert (
        str(exc.value)
        .strip()
        .startswith("Calling add_task: Arguments must _always_ be named")
    )


def test_nesting_non_subworkflows_throws_error() -> None:
    """Ensure nesting is only allow in subworkflows.

    Nested tasks must be evaluated at construction time, and there
    is no concept of task calls that are not resolved during construction, so
    a task should not be called inside a non-nested task.
    """
    with pytest.raises(TaskException) as exc:
        badly_wrap_task()
    assert (
        str(exc.value)
        .strip()
        .startswith(
            "You referenced a task add_task inside another task badly_wrap_task, but it is not a workflow"
        )
    )


def test_nesting_does_not_identify_imports_as_nesting() -> None:
    """Ensure we do not throw errors simply because a task's name appears in an import.

    Non-nested tasks should error if a task appears in the body, but due to https://bugs.python.org/issue36697
    this can be misidentified using getclosurevars.

    TODO: The invisible vs visible difference below is an example of a known bug.
    It seems better to have false negatives than positives, but if an import has not
    already put the module into the sys.modules, then we cannot spot it with this workaround.
    The hidden-by-math examples are also wrong - by inspect module alone, we are not sure if `math` is an import
    and (even if it's not _too_ hard to work that out) whether the variable pi actually comes from it.

    UPDATE (2025/01/18): The visible version strictly succeeded by a mischaracterization
    of getclosurevars, where an imported symbol was conflated with a global that shared its name.
    Unfortunately, addressing that (as 3.12.8 does) makes it harder to spot incorrect task use from modules,
    but we do not wish to execute an inline import to confirm either way, not knowing if the module is
    available or would otherwise be imported during the rendering. A potential improvement would at least
    check whether the module has been loaded and look, and the logic required to link an imported symbol
    to the full module path (in most circumstances) is now in FunctionAnalyser.unbound. The next step
    would be to assess the reliability of matching those modules to existing imports, and the footgun of
    false negatives when the module has never been imported outside a task (or before hitting that definition).

    One direction to go with this would be to see how mypy follows this during inspection and see if we could
    take the same approach.
    """
    good = [
        pi,
        pi_exported_from_math,
        pi_with_invisible_module_task,
        pi_with_visible_module_task,
        pi_hidden_by_math,
        pi_hidden_by_math_2,
    ]
    bad = [try_recursive]
    for tsk in bad:
        with pytest.raises(TaskException) as exc:
            tsk()
        assert str(exc.value).strip().startswith("You referenced a task")
    for tsk in good:
        result = tsk()
        construct(result)


def test_normal_objects_cannot_be_used_in_subworkflows() -> None:
    """Most entities cannot appear in a subworkflow, ensure we catch them.

    Since the logic in nested tasks has to be embedded explicitly in the workflow,
    complex types are not necessarily representable, and in most cases, we would not
    be able to guarantee that the libraries, versions, etc. match.

    Note: this may be mitigated with sympy support, to some extent.
    """
    with pytest.raises(TaskException) as exc:
        unacceptable_object_usage()
    assert (
        str(exc.value)
        == "Attempted to build a workflow from a return-value/result/expression with no references."
    )


def test_subworkflows_must_return_a_task() -> None:
    """Ensure nested tasks are lazy-evaluatable.

    A graph only makes sense if the edges connect, and nested tasks must therefore chain.
    As such, a nested task must represent a real subgraph, and return a node to pull it into
    the main graph.
    """
    with pytest.raises(TaskException) as exc:
        result = unacceptable_nested_return(int_not_global=True)
        construct(result)
    assert (
        str(exc.value)
        == "Attempted to build a workflow from a return-value/result/expression with no references."
    )

    result = unacceptable_nested_return(int_not_global=False)
    construct(result)


bad_num = 3
good_num: int = 4


def test_must_annotate_global() -> None:
    """TODO: Docstrings."""
    worse_num = 3

    @workflow()
    def check_annotation() -> int | float:
        return increment(num=bad_num)

    with pytest.raises(TaskException) as exc:
        check_annotation()

    assert (
        str(exc.value)
        == "Could not find a type annotation for bad_num for check_annotation"
    )

    @workflow()
    def check_annotation_2() -> int | float:
        return increment(num=worse_num)

    with pytest.raises(TaskException) as exc:
        check_annotation_2()

    assert (
        str(exc.value)
        == "Cannot use free variables - please put worse_num at the global scope"
    )

    @workflow()
    def check_annotation_3() -> int | float:
        return increment(num=good_num)

    check_annotation_3()



# Copyright 2014 Flax & Teal Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# with WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Overarching workflow concepts.

Basic constructs for describing a workflow.
"""

from __future__ import annotations
import inspect
from collections.abc import Mapping, MutableMapping, Callable
from attrs import has as attr_has, resolve_types, fields as attrs_fields
from dataclasses import is_dataclass, fields as dataclass_fields
from collections import Counter, OrderedDict
from typing import (
    Protocol,
    Any,
    ParamSpec,
    TypeVar,
    Generic,
    cast,
    Literal,
    Iterable,
    get_origin,
    get_args,
    Generator,
    Sized,
    Sequence,
    get_type_hints,
    TYPE_CHECKING,
    Hashable,
)
from uuid import uuid4
import logging

from sympy import Symbol, Expr, Basic, Tuple

logger = logging.getLogger(__name__)

from .core import (
    IterableMixin,
    Reference,
    get_configuration,
    Raw,
    IteratedGenerator,
    strip_annotations,
    WorkflowProtocol,
    WorkflowComponent,
    ExprType,
)
from .utils import hasher, is_raw, make_traceback, is_raw_type, is_expr, Unset

T = TypeVar("T")
U = TypeVar("U")
RetType = TypeVar("RetType")
Param = ParamSpec("Param")

CHECK_IDS = False
AVAILABLE_TYPES = {"int": int, "str": str}


class Lazy(Protocol):
    """Requirements for a lazy-evaluatable function."""

    __name__: str

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """When called this should return a reference."""
        ...


class LazyEvaluation(Lazy, Generic[RetType]):
    """Tracks a single evaluation of a lazy function."""

    def __init__(self, fn: Callable[Param, RetType]):
        """Initialize an evaluation.

        Args:
            fn: callable returning RetType, which this will return
                also from it's __call__ method for consistency.
        """
        self._fn: Callable[Param, RetType] = fn
        self.__name__ = fn.__name__

    def __call__(self, *args: Any, **kwargs: Any) -> RetType:
        """Wrapper around a lazy execution.

        Captures a traceback, for debugging if this does not work.

        WARNING: this is one of the few places that we would expect
        dask distributed to break, if running outside a single process
        is attempted.
        """
        tb = make_traceback()
        result = self._fn(*args, **kwargs, __traceback__=tb)  # type: ignore[arg-type]
        return result


Target = Callable[..., Any]
StepExecution = Callable[..., Lazy]
LazyFactory = Callable[[Target], Lazy]


class UnsetType(Unset, Generic[T]):
    """Unset variable with a specific type.

    Attributes:
        __type__: type of the variable.
    """

    __type__: type[T]

    def __init__(self, raw_type: type[T]):
        """Create a new Unset token of a specific type.

        Attributes:
            __type__: type of the variable.
        """
        self.__type__ = raw_type


UNSET = Unset()


class Parameter(Generic[T], Symbol):
    """Global parameter.

    Independent parameter that will be used when a task is spotted
    reaching outside its scope. This wraps the variable it uses.

    To allow for potential arithmetic operations, etc. it is a Sympy
    symbol.

    Attributes:
        __name__: name of the parameter.
        __default__: captured default value from the original value.
    """

    __name__: str
    __name_suffix__: str = ""
    __default__: T | UnsetType[T]
    __tethered__: Literal[False] | None | BaseStep | Workflow
    __fixed_type__: type[T] | Unset
    autoname: bool = False

    def __init__(
        self,
        name: str,
        default: T | UnsetType[T],
        tethered: Literal[False] | None | Step | Workflow = None,
        autoname: bool = False,
        typ: type[T] | Unset = UNSET,
    ):
        """Construct a parameter.

        Args:
            name: name of the parameter.
            default: value to infer type, etc. from.
            tethered: a workflow or step that demands this parameter; None if not yet present, False if not desired.
            autoname: whether we should customize this name for uniqueness (it is not user-set).
            typ: a type to override what would be automatically inferred.
        """
        self.__original_name__ = name

        # TODO: is using this in a step hash a risk of ambiguity? (full name is circular)
        if autoname:
            self.__name_suffix__ = f"-{uuid4()}"
        self.autoname = autoname

        self.__name__ = name
        self.__default__ = default
        self.__tethered__ = tethered
        self.__callers__: list[BaseStep] = []
        self.__fixed_type__ = typ

        if tethered and isinstance(tethered, BaseStep):
            self.register_caller(tethered)

    @staticmethod
    def is_loopable(typ: type) -> bool:
        """Checks if this type can be looped over.

        In particular, checks if this is an iterable that is NOT a str or bytes, possibly disguised
        behind an Annotated.

        Args:
            typ: type to check.

        Returns: True if loopable, otherwise False.
        """
        base = strip_annotations(typ)[0]
        base = get_origin(base) or base
        return (
            inspect.isclass(base)
            and issubclass(base, Iterable)
            and not issubclass(base, str | bytes)
        )

    @property
    def __type__(self) -> type | Unset:
        """Type associated with this parameter."""
        if self.__fixed_type__ is not UNSET:
            return self.__fixed_type__

        default = self.__default__
        if (
            default is not None
            and hasattr(default, "__type__")
            and isinstance(default.__type__, type)
        ):
            raw_type = default.__type__
        else:
            raw_type = type(default)
        return raw_type

    def __eq__(self, other: object) -> bool:
        """Comparing two parameters.

        Currently, this uses the hashes.

        TODO: confirm this is an iff.
        """
        return hash(self) == hash(other)

    def __new__(cls, *args: Any, **kwargs: Any) -> "Parameter[T]":
        """Creates a Parameter.

        Required, as Parameters are an instance of sympy Expression, so
        we must instantiate via it.
        """
        instance = Expr.__new__(cls)
        instance._assumptions0 = {}
        return cast(Parameter[T], instance)

    def __hash__(self) -> int:
        """Get a unique hash for this parameter."""
        # if self.__tethered__ is None:
        #     raise RuntimeError(
        #         f"Parameter {self.name} was never tethered but should have been"
        #     )
        return hash(self.__name__)

    def make_reference(self, **kwargs: Any) -> "ParameterReference[T]":
        """Creates a new reference for the parameter.

        The kwargs will be passed to the constructor, but the

        Args:
            typ: type of the new reference's target.
            **kwargs: arguments to pass to the constructor.

        Returns: checks if `typ` is loopable and gives an IterableParameterReference if so,
            otherwise a normal ParameterReference.
        """
        kwargs["parameter"] = self
        kwargs.setdefault("typ", self.__type__)
        typ = kwargs["typ"]
        if self.is_loopable(typ):
            return IterableParameterReference(**kwargs)
        return ParameterReference(**kwargs)

    @property
    def default(self) -> T | UnsetType[T]:
        """Retrieve default value for this parameter, or an unset token."""
        return self.__default__

    @property
    def name(self) -> str:
        """Extended name, suitable for rendering.

        This attempts to create a unique name by tying the parameter to a step
        if the user has not explicitly provided a name, ideally the one where
        we discovered it.
        """
        tethered = self.__tethered__
        if tethered is False or tethered is None or self.autoname is False:
            return self.__name__ + self.__name_suffix__
        else:
            return f"{tethered.name}-{self.__original_name__}"

    def register_caller(self, caller: BaseStep) -> None:
        """Capture a step that uses this parameter.

        Gathers together the steps using this parameter. The first found will
        be recorded as the tethered step, and used for forming the name.
        """
        if self.__tethered__ is None:
            self.__tethered__ = caller
        self.__callers__.append(caller)

    def __getattr__(self, attr: str) -> Reference[T] | Any:
        """Retrieve a reference to a field within this Parameter.

        Arg:
            attr: a field to find.

        Returns: a new reference with the `attr` appended to the field tuple.
        """
        return getattr(self.make_reference(workflow=None), attr)


def param(
    name: str,
    default: T | UnsetType[T] | Unset = UNSET,
    tethered: Literal[False] | None | Step | Workflow = False,
    typ: type[T] | Unset = UNSET,
    autoname: bool = False,
) -> T:
    """Create a parameter.

    Will cast so it looks like the original type.

    Returns:
        Parameter class cast to the type of the supplied default.
    """
    if default is UNSET:
        if isinstance(typ, Unset):
            raise ValueError("Must provide a default or a type")
        default = UnsetType[T](typ)
    return cast(
        T,
        Parameter(name, default=default, tethered=tethered, autoname=autoname, typ=typ),
    )


class Task:
    """Named wrapper of a lazy-evaluatable function.

    Wraps a lazy-evaluatable function (`dewret.workflow.Lazy`) with any
    metadata needed to render it later. At present, this is the name.

    Attributes:
        name: Name of the lazy function as it will appear in the output workflow text.
        target: Callable that is wrapped.
    """

    name: str
    target: Lazy

    def __init__(self, name: str, target: Lazy):
        """Initialize the Task.

        Args:
            name: Name of wrapped function.
            target: Actual function being wrapped (optional).
        """
        self.name = name
        self.target = target

    @property
    def __name__(self) -> str:
        """Name of the task."""
        return self.name

    def __str__(self) -> str:
        """Stringify the Task, currently by returning the `name`."""
        return self.name

    def __repr__(self) -> str:
        """Represent the Task, currently by returning the `name`."""
        return self.name

    def __hash__(self) -> int:
        """Hashes for finding."""
        return hash(self.name)

    def __eq__(self, other: object) -> bool:
        """Is this the same task?

        At present, we naively compare the name and target. In
        future, this may be more nuanced.
        """
        if not isinstance(other, Task):
            return False
        return self.name == other.name and self.target == other.target


class Workflow:
    """Overarching workflow concept.

    Represents a whole workflow, as a singleton maintaining all
    state information needed ahead of rendering. It is built up
    as the lazy-evaluations are finally evaluated.

    Attributes:
        steps: the sequence of calls to lazy-evaluable functions,
            built as they are evaluated.
        tasks: the mapping of names used in the `steps` to the actual
            `Task` wrappers they represent.
        result: target reference to evaluate, if yet present.
    """

    _steps: list["BaseStep"]
    tasks: MutableMapping[str, "Task"]
    result: (
        StepReference[Any]
        | list[StepReference[Any]]
        | tuple[StepReference[Any], ...]
        | None
    )
    _remapping: dict[str, str] | None
    _name: str | None

    def __init__(self, name: str | None = None) -> None:
        """Initialize a Workflow, by setting `steps` and `tasks` to empty containers."""
        self._steps = []
        self.tasks = {}
        self.result: StepReference[Any] | None = None
        self._remapping = None
        self._name = name

    @property
    def steps(self) -> set[BaseStep]:
        """Get deduplicated steps.

        Returns: steps for looping over without duplicates.
        """
        return set(self._steps)

    def __str__(self) -> str:
        """Name of the workflow, if available."""
        if self._name is None:
            return super().__str__()
        return self.name

    def __repr__(self) -> str:
        """Representation of the workflow.

        This will be the `id` if the workflow is anonymous.

        Returns: string identifier for this workflow.
        """
        if self._name:
            return self.name
        return self.id

    def __hash__(self) -> int:
        """Hashes for finding."""
        return hash(self.id)

    @property
    def id(self) -> str:
        """Consistent ID based off the step IDs."""
        comp_tup = tuple(sorted(s.id for s in self.steps))
        return f"workflow-{hasher(comp_tup)}"

    def __eq__(self, other: object) -> bool:
        """Is this the same workflow?

        At present, we naively compare the steps and arguments. In
        future, this may be more nuanced.
        """
        if not isinstance(other, Workflow):
            return False
        return (
            self._steps == other._steps
            and self.tasks == other.tasks
            and self.result == other.result
            and self._remapping == other._remapping
            and self._name == other._name
        )

    @property
    def has_result(self) -> bool:
        """Confirms whether this workflow has a non-empty result.

        Either None or an empty list/tuple are considered empty for this purpose.

        Returns: True if the workflow has a result, False otherwise.
        """
        return not (
            self.result is None
            or (isinstance(self.result, list | tuple) and not self.result)
        )

    @property
    def name(self) -> str:
        """Get the name of the workflow.

        Raises:
            NameError: if no name has been set.
        """
        if self._name is None:
            raise NameError("Can not get the name of an anonymous workflow.")
        return self._name

    def find_factories(self) -> dict[str, FactoryCall]:
        """Steps that are factory calls."""
        return {
            step_id: step
            for step_id, step in self.indexed_steps.items()
            if isinstance(step, FactoryCall)
        }

    def find_parameters(
        self, include_factory_calls: bool = True
    ) -> set[Parameter[Any]]:
        """Crawl steps for parameter references.

        As the workflow does not hold its own list of parameters, this
        dynamically finds them.

        Returns:
            Set of all references to parameters across the steps.
        """
        _, references = expr_to_references(
            step.arguments
            for step in self.steps
            if (include_factory_calls or not isinstance(step, FactoryCall))
        )
        return {
            ref._.parameter for ref in references if isinstance(ref, ParameterReference)
        }

    @property
    def indexed_steps(self) -> dict[str, BaseStep]:
        """Steps mapped by ID.

        Forces generation of IDs. Note that this effectively
        freezes the steps, so it should not be used until we
        are confident the steps are all ready to be hashed.

        Returns:
            Mapping of steps by ID.
        """
        return OrderedDict(
            sorted(((step.id, step) for step in self.steps), key=lambda x: x[0])
        )

    @classmethod
    def assimilate(cls, *workflow_args: Workflow) -> "Workflow":
        """Combine two Workflows into one Workflow.

        Takes two workflows and unifies them by combining steps
        and tasks. If it sees mismatched identifiers for the same
        component, it will error.
        This could happen if the hashing function is flawed
        or some Python magic to do with Targets being passed.

        Args:
            workflow_args: workflows to use as base

        j
        """
        workflows = sorted((w for w in set(workflow_args)), key=lambda w: w.id)
        base = workflows[0]

        if len(workflows) == 1:
            return base

        names = sorted({w._name for w in workflows if w._name})
        base._name = base._name or (names and names[0]) or None

        # left_steps = left._indexed_steps
        # right_steps = right._indexed_steps
        all_steps = sorted(
            sum((list(w.indexed_steps.items()) for w in workflows), []),
            key=lambda s: s[0],
        )

        for _, step in all_steps:
            # for step in list(left_steps.values()) + list(right_steps.values()):
            step.set_workflow(base)

        indexed_steps: dict[str, BaseStep] = {}
        for step_id, step in all_steps:
            indexed_steps.setdefault(step_id, step)
            if step != indexed_steps[step_id]:
                raise RuntimeError(
                    f"Two steps have same ID but do not match: {step_id}"
                )

        all_tasks = sum((list(w.tasks.items()) for w in workflows), [])
        indexed_tasks: dict[str, Task] = {}
        for task_id, task in all_tasks:
            indexed_tasks.setdefault(task_id, task)
            if task != indexed_tasks[task_id]:
                raise RuntimeError(
                    f"Two tasks have same name {task_id} but do not match"
                )

        base._steps = sorted(indexed_steps.values(), key=lambda s: s.id)
        base.tasks = indexed_tasks

        for step in base.steps:
            step.set_workflow(base, with_arguments=True)

        hashable_workflows: list[Workflow] = [
            w for w in workflows if isinstance(w.result, Hashable)
        ]
        if len(hashable_workflows) != len(workflows):
            raise NotImplementedError("Some results are not hashable.")

        def _get_order(
            result: None | StepReference[Any] | Iterable[StepReference[Any]],
        ) -> str:
            if result is None:
                return ""
            if isinstance(result, StepReference):
                return result.id
            return "|".join(r for r in result)

        results = sorted(
            set({w.result for w in hashable_workflows if w.has_result}),
            key=lambda r: _get_order(r),
        )
        if len(results) == 1:
            result = results[0]
        else:
            list_results = [
                r if isinstance(r, tuple | list | Tuple) else (r,) for r in results
            ]
            result = sum(map(list, list_results), [])

        if result is not None and result != []:
            unify_workflows(result, base, set_only=True)
            base.set_result(result)

        return base

    def remap(self, step_id: str) -> str:
        """Apply name simplification if requested.

        Args:
            step_id: step to check.

        Returns:
            Same ID or a remapped name.
        """
        return self._remapping.get(step_id, step_id) if self._remapping else step_id

    def simplify_ids(self, infix: list[str] | None = None) -> None:
        """Work out mapping to simple ints from hashes.

        Goes through and numbers each step by the order of use of its task.
        """
        counter = Counter[Task | Workflow]()
        self._remapping = {}
        infix_str = ("-".join(infix) + "-") if infix else ""
        for step in self.indexed_steps.values():
            counter[step.task] += 1
            self._remapping[step.id] = f"{step.task}-{infix_str}{counter[step.task]}"
            if isinstance(step, NestedStep):
                step.subworkflow.simplify_ids(
                    infix=(infix or []) + [str(counter[step.task])]
                )
        param_counter = Counter[str]()
        name_to_original: dict[str, str] = {}
        for name, param in {
            param.__name__: param
            for param in self.find_parameters()
            if isinstance(param, Parameter)
        }.items():
            if param.__original_name__ != name:
                param_counter[param.__original_name__] += 1
                self._remapping[param.__name__] = (
                    f"{param.__original_name__}-{param_counter[param.__original_name__]}"
                )
                name_to_original[param.__original_name__] = param.__name__
        for pname, count in param_counter.items():
            if count == 1:
                self._remapping[name_to_original[pname]] = pname

    def register_task(self, fn: Lazy) -> Task:
        """Note the existence of a lazy-evaluatable function, and wrap it as a `Task`.

        Args:
            fn: the wrapped function.

        Returns:
            A new `Task` that wraps the function, and is retained in the `Workflow.tasks`
            dict.
        """
        name = fn.__name__
        if name in self.tasks and self.tasks[name].target != fn:
            raise RuntimeError(f"Naming clash for functions: {name}")

        task = Task(name, fn)
        self.tasks[name] = task
        return task

    def add_nested_step(
        self,
        name: str,
        subworkflow: Workflow,
        return_type: type | None,
        kwargs: dict[str, Any],
        positional_args: dict[str, bool] | None = None,
    ) -> StepReference[Any]:
        """Append a nested step.

        Calls a subworkflow.

        Args:
            name: name of the subworkflow.
            subworkflow: the subworkflow itself.
            return_type: a forced type for the return, or None.
            kwargs: any key-value arguments to pass in the call.
            positional_args: a mapping of arguments to bools, True if the argument is positional or otherwise False.

        Returns: a reference to the step that calls out to a new workflow.
        """
        step = NestedStep(self, name, subworkflow, kwargs)
        if positional_args is not None:
            step.positional_args = positional_args
        self._steps.append(step)
        return_type = return_type or step.return_type
        if return_type is inspect._empty:
            raise TypeError("All tasks should have a type annotation.")
        return step.make_reference(workflow=self, typ=return_type)

    def add_step(
        self,
        fn: Lazy,
        kwargs: dict[str, Raw | Reference[Any]],
        raw_as_parameter: bool = False,
        is_factory: bool = False,
        positional_args: dict[str, bool] | None = None,
    ) -> StepReference[Any]:
        """Append a step.

        Adds a step, for running a target with key-value arguments,
        to the workflow.

        Args:
            fn: the target function to turn into a step.
            kwargs: any key-value arguments to pass in the call.
            raw_as_parameter: whether to turn any discovered raw arguments into workflow parameters.
            is_factory: whether this step should be a Factory.
            positional_args: a mapping of arguments to bools, True if the argument is positional or otherwise False.
        """
        task = self.register_task(fn)
        step_maker = FactoryCall if is_factory else Step
        step = step_maker(self, task, kwargs, raw_as_parameter=raw_as_parameter)
        if positional_args is not None:
            step.positional_args = positional_args
        self._steps.append(step)
        return_type = step.return_type
        if (
            return_type is inspect._empty
            and not isinstance(fn, type)
            and not inspect.isclass(fn)
        ):
            raise TypeError("All tasks should have a type annotation.")
        return step.make_reference(workflow=self, typ=return_type)

    @staticmethod
    def from_result(
        result: StepReference[Any]
        | list[StepReference[Any]]
        | tuple[StepReference[Any], ...],
        simplify_ids: bool = False,
        nested: bool = True,
    ) -> Workflow:
        """Create from a desired result.

        Starts from a result, and builds a workflow to output it.
        """
        expr, _refs = expr_to_references(result)
        if not _refs:
            raise RuntimeError(
                "Attempted to build a workflow from a return-value/result/expression with no references."
            )
        refs = list(_refs)
        if isinstance(refs[0], Parameter):
            raise RuntimeError("Attempted to build a workflow from an input parameter.")
        workflow = refs[0].__workflow__
        # Ensure that we have exactly one workflow, even if multiple results.
        for entry in refs[1:]:
            if entry.__workflow__ != workflow:
                raise RuntimeError(
                    "If multiple results, they must share a single workflow"
                )
        workflow.set_result(expr)
        if simplify_ids:
            workflow.simplify_ids()
        if not isinstance(workflow, Workflow):
            # This may be better as a cast, since this seems an artificial case.
            raise TypeError("Can only return a workflow of the same type")
        return workflow

    def set_result(self, result: Basic | list[Basic] | tuple[Basic]) -> None:
        """Choose the result step.

        Sets a step as being the result for the entire workflow.
        When we evaluate a dynamic workflow, the engine (e.g. dask)
        creates a graph to realize the result of a single collection.
        Similarly, in the static case, we need to have a result that
        drives the calculation.

        Args:
            result: reference to the chosen step.
        """
        _, refs = expr_to_references(result)
        for entry in refs:
            if entry.__workflow__ != self:
                raise RuntimeError("Output must be from a step in this workflow.")
        self.result = result

    @property
    def result_type(self) -> type:
        """Overall return type of this workflow."""
        if self.result is None:
            return type(None)
        if hasattr(self.result, "__type__"):
            return self.result.__type__
        # TODO: get individual types!
        return type(self.result)


class WorkflowLinkedComponent(Protocol):
    """Protocol for objects dynamically tied to a `Workflow`."""

    @property
    def __workflow__(self) -> Workflow:
        """Workflow currently tied to.

        Usually a proxy for another object that it should
        consistently follow.

        Returns:
            workflow: the `Workflow` to tie to.
        """
        ...


class FieldableProtocol(Protocol):
    """Expected interfaces for a type that can take fields.

    Attributes:
        __field__: tuple representing the named fields, either strings or integers.
    """

    __field__: tuple[str | int, ...]

    @property
    def __field_sep__(self) -> str:
        """The separator that should be used by default when rendering a name."""

    @property
    def __field_index_types__(self) -> tuple[type, ...]:
        """Get the types that will be rendered as an index [x] rather than a field .x on stringifying.

        Will be taken from the `field_index_types` construct configuration.
        """

    def __init__(self, *args: Any, field: str | None = None, **kwargs: Any):
        """Extract the field name from the initializer arguments, if provided."""
        super().__init__(*args, **kwargs)

    @property
    def __name__(self: FieldableProtocol) -> str:
        """Name of the fieldable, which may not be `name` if this is needed as an internal reference."""
        return self.name

    @property
    def __type__(self) -> type:
        """Type of this field with the overall target."""
        ...

    @property
    def name(self) -> str:
        """The name for the target, accounting for the field."""
        return "name"

    def __make_reference__(self, *args: Any, **kwargs: Any) -> Reference[Any]:
        """Create a reference with constructor arguments, usually to a subfield."""
        ...


# This required so that Mixins can trust that
# superclasss attributes will be available for mypy,
# but do not confuse the MRO.
if TYPE_CHECKING:
    _Fieldable = FieldableProtocol
else:
    _Fieldable = object


# Subclass Reference so that we know Reference methods/attrs are available.
class FieldableMixin(_Fieldable):
    """Tooling for enhancing a type with referenceable fields."""

    def __init__(
        self: FieldableProtocol,
        *args: Any,
        field: str | int | tuple[str | int, ...] | None = None,
        **kwargs: Any,
    ):
        """Extract the requested field, if any, from the initializer arguments.

        Args:
            field: the new subfield, either an index (int) or a fieldname (str) or a field tuple representing the whole path, or None.
            *args: arguments to pass to the other initializers.
            **kwargs: arguments to pass to the other initializers.
        """
        self.__field__: tuple[str, ...] = (
            (field if isinstance(field, tuple) else (field,))
            if field is not None
            else ()
        )
        super().__init__(*args, **kwargs)

    @property
    def __field_sep__(self) -> str:
        """Get the field separator.

        Will be taken from the configuration key, `field_separator`.
        """
        sep = get_configuration("field_separator")
        if not isinstance(sep, str):
            raise TypeError(
                f"The `field_separator` configuration argument must be a string not {type(sep)}"
            )
        return sep

    @property
    def __field_index_types__(self) -> tuple[type, ...]:
        """Get the types that will be rendered as an index [x] rather than a field .x on stringifying.

        Will be taken from the `field_index_types` construct configuration.
        """
        types = get_configuration("field_index_types")
        if not isinstance(types, str):
            raise TypeError(
                f"The `field_index_types` configuration argument should be a comma-separated string of types from: {', '.join(AVAILABLE_TYPES.keys())}"
            )
        tup_all = tuple(AVAILABLE_TYPES.get(typ) for typ in types.split(",") if typ)
        tup = tuple(t for t in tup_all if t is not None)
        if tup_all != tup:
            raise ValueError(
                "Setting for fixed index types contains unavailable type: "
                + f"{str(get_configuration('field_index_types'))} vs {tup}"
            )
        return tup

    @property
    def __field_str__(self) -> str:
        """Stringified field, without the name."""
        return self.__field_suffix__.lstrip(self.__field_sep__)

    @property
    def __field_suffix__(self) -> str:
        """Stringified field, without the name but with an appropriate separator at the start."""
        result = ""
        for cmpt in self.__field__:
            if any(isinstance(cmpt, typ) for typ in self.__field_index_types__):
                result += f"[{cmpt}]"
            else:
                result += f"{self.__field_sep__}{cmpt}"
        return result

    def __field_up__(self) -> Reference[Any]:
        """Get the parent field, if possible.

        Returns: reference to the field above this.

        Raises:
            RuntimeError: if there is no field above this.
        """
        if self.__field__:
            return self.__make_reference__(field=tuple(self.__field__[:-1]))
        raise RuntimeError("Cannot go upwards unless currently using a field.")

    @property
    def __name__(self) -> str:
        """Name for this step.

        May be remapped by the workflow to something nicer
        than the ID.
        """
        return super().__name__ + self.__field_suffix__

    def find_field(
        self: FieldableProtocol,
        field: str | int,
        fallback_type: type | None = None,
        **init_kwargs: Any,
    ) -> Reference[Any]:
        """Field within the reference, if possible.

        Args:
            field: the field to search for.
            fallback_type: the type to use if we do not know a more specific one.
            **init_kwargs: arguments to use for constructing a new reference (via `__make_reference__`).

        Returns:
            A field-specific version of this reference.
        """
        # Get new type, for the specific field.
        parent_type, _ = strip_annotations(self.__type__)
        field_type = fallback_type
        base = get_origin(parent_type) or parent_type

        if isinstance(field, int):
            if not inspect.isclass(base) or not issubclass(base, Sequence):
                raise AttributeError(
                    f"Tried to index int {field} into a non-sequence type {parent_type} (base: {base})"
                )
            if not (field_type := get_args(parent_type)[0]):
                raise AttributeError(
                    f"Tried to index int {field} into type {parent_type} but can only do so if the first type argument "
                    f"is the element type (args: {get_args(parent_type)}"
                )
        elif field.startswith("__"):
            raise AttributeError(
                f"We do not allow fields with dunder prefix, such as {field}, to reduce risk of clashes."
            )
        else:
            if is_dataclass(parent_type):
                try:
                    type_hints = get_type_hints(
                        parent_type,
                        localns={parent_type.__name__: parent_type},
                        include_extras=True,
                    )
                    field_type = type_hints.get(field)
                    if field_type is None:
                        dataclass_field_type = next(
                            iter(
                                filter(
                                    lambda fld: fld.name == field,
                                    dataclass_fields(parent_type),
                                )
                            )
                        ).type
                        if isinstance(dataclass_field_type, str):
                            # TODO: we could ask Python to resolve the str expression for us
                            raise TypeError(
                                "Dataclass fields must be provided as types directly, not str"
                            )
                        field_type = dataclass_field_type
                except StopIteration:
                    raise AttributeError(
                        f"Dataclass {parent_type} does not have field {field}"
                    ) from None
            elif attr_has(parent_type):
                resolve_types(parent_type)
                try:
                    field_type = getattr(attrs_fields(parent_type), field).type
                except AttributeError as exc:
                    raise AttributeError(
                        f"attrs-class {parent_type} does not have field {field}"
                    ) from exc
            # TypedDict
            elif (
                inspect.isclass(parent_type)
                and issubclass(parent_type, dict)
                and hasattr(parent_type, "__annotations__")
            ):
                try:
                    field_type = get_type_hints(parent_type, include_extras=True)[field]
                except KeyError as exc:
                    raise AttributeError(
                        f"TypedDict {parent_type} does not have field {field}"
                    ) from exc
            if (
                not field_type
                and get_configuration("allow_plain_dict_fields")
                and inspect.isclass(base)
                and issubclass(base, dict)
            ):
                args = get_args(parent_type)
                if len(args) == 2 and args[0] is str:
                    field_type = args[1]
                else:
                    raise AttributeError(
                        "Can only get fields for plain dicts if annotated dict[str, TYPE]"
                    )

        if field_type:
            if not issubclass(self.__class__, Reference):
                raise TypeError("Only references can have a fieldable mixin")

            new_field: tuple[str | int, ...] | str | int
            if self.__field__:
                new_field = tuple(list(self.__field__) + [field])
            else:
                new_field = field

            return self.__make_reference__(
                typ=field_type, field=new_field, **init_kwargs
            )

        raise AttributeError(
            f"Could not determine the type for field {field} in type {parent_type} (type of parent type is {type(parent_type)})"
        )


class BaseStep(WorkflowComponent):
    """Lazy-evaluated function call.

    Individual function call to a lazy-evaluatable function, tracked
    for building up the `Workflow`.

    Attributes:
        task: the `Task` being called in this step.
        arguments: key-value pairs of arguments to this step.
    """

    _id: str | None = None
    task: Task | Workflow
    arguments: Mapping[str, Basic | Reference[Any] | Raw]
    workflow: Workflow
    positional_args: dict[str, bool] | None = None

    def __init__(
        self,
        workflow: Workflow,
        task: Task | Workflow,
        arguments: Mapping[str, Reference[Any] | Raw],
        raw_as_parameter: bool = False,
    ):
        """Initialize a step.

        Args:
            workflow: `Workflow` that this is tied to.
            task: the lazy-evaluatable function that this wraps.
            arguments: key-value pairs to pass to the function.
            raw_as_parameter: whether to turn any raw-type arguments into workflow parameters (or just keep them as default argument values).
        """
        super().__init__(workflow=workflow)
        self.task = task
        self.arguments = {}
        for key, value in arguments.items():
            if (
                isinstance(value, FactoryCall)
                or isinstance(value, Reference)
                or isinstance(value, Raw)
                or is_raw(value)
                or is_expr(value)
                or is_dataclass(value)
                or attr_has(value)
            ):
                # Avoid recursive type issues
                if (
                    not isinstance(value, Reference)
                    and not isinstance(value, FactoryCall)
                    and not isinstance(value, Raw)
                    and is_raw(value)
                ):
                    if raw_as_parameter:
                        # We use param for convenience but note that it is a reference in disguise.
                        value = cast(
                            Parameter[Any], param(key, value, tethered=None)
                        ).make_reference(workflow=workflow)
                    else:
                        value = Raw(value)

                def _to_param_ref(value: Any) -> ParameterReference[Any] | None:
                    if isinstance(value, Parameter):
                        return value.make_reference(workflow=workflow)
                    return None

                expression, refs = expr_to_references(value, remap=_to_param_ref)

                for ref in refs:
                    if isinstance(ref, ParameterReference):
                        parameter = ref._.parameter
                        parameter.register_caller(self)
                self.arguments[key] = expression
            else:
                raise RuntimeError(
                    f"Non-references must be a serializable type: {key}>{value} {type(value)}"
                )

    def __eq__(self, other: object) -> bool:
        """Is this the same step?

        At present, we naively compare the task and arguments. In
        future, this may be more nuanced.
        """
        if not isinstance(other, BaseStep):
            return False
        return (
            self.__workflow__ is other.__workflow__
            and self.task == other.task
            and self.arguments == other.arguments
        )

    def make_reference(self, **kwargs: Any) -> "StepReference[T]":
        """Create a reference to this step.

        Builds a reference to the (result of) this step, which will be iterable if appropriate.

        Args:
            **kwargs: arguments for reference constructor, which will be supplemented appropriately.

        Returns: a reference to the result of this step.
        """
        kwargs["step"] = self
        kwargs.setdefault("typ", self.return_type)
        typ = kwargs["typ"]
        base = strip_annotations(typ)[0]
        base = get_origin(base) or base
        if (
            inspect.isclass(base)
            and issubclass(base, Iterable)
            and not issubclass(base, str | bytes)
        ):
            return IterableStepReference(**kwargs)
        return StepReference(**kwargs)

    def __hash__(self) -> int:
        """Searchable hash for this step."""
        return hash(self.id)

    def set_workflow(self, workflow: Workflow, with_arguments: bool = True) -> None:
        """Move the step reference to another workflow.

        This method is primarily intended to be called by a step, allowing it to
        switch to a new workflow. It also updates the workflow reference for any
        arguments that are steps themselves, if specified.

        Args:
            workflow: The new target workflow to which the step should be moved.
            with_arguments: If True, also update the workflow reference for the step's arguments.
        """
        self.__workflow__ = workflow
        if with_arguments:
            for argument in self.arguments.values():
                unify_workflows(argument, workflow, set_only=True)
        self._id = None

    @property
    def return_type(self) -> Any:
        """Take the type of the wrapped function from the target.

        Unwraps and inspects the signature, meaning that the original
        wrapped function _must_ have a typehint for the return value.

        Returns:
            Expected type of the return value.
        """
        if isinstance(self.task, Workflow):
            if self.task.result is not None:
                return self.task.result_type
            else:
                raise AttributeError(
                    "Cannot determine return type of a workflow with an unspecified result"
                )
        if isinstance(self.task.target, type):
            return self.task.target
        ann = get_type_hints(inspect.unwrap(self.task.target), include_extras=True)[
            "return"
        ]
        return ann

    @property
    def name(self) -> str:
        """Name for this step.

        May be remapped by the workflow to something nicer
        than the ID.
        """
        return self.__workflow__.remap(self.id)

    @property
    def id(self) -> str:
        """Consistent ID based on the value."""
        if self._id is None:
            self._id = self._generate_id()
            return self._id

        if CHECK_IDS:
            check_id = self._generate_id()
            if check_id != self._id:
                return self._id
                raise RuntimeError(
                    f"Cannot change a step after requesting its ID: {self.task}"
                )
        return self._id

    def _generate_id(self) -> str:
        """Generate the ID once."""
        components: list[str | tuple[str, str]] = [repr(self.task)]
        for key, param in self.arguments.items():
            components.append((key, repr(param)))

        comp_tup: tuple[str | tuple[str, str], ...] = tuple(
            sorted(components, key=lambda pair: pair[0])
        )

        return f"{self.task}-{hasher(comp_tup)}"


class NestedStep(BaseStep):
    """Calling out to a subworkflow.

    Type of BaseStep to call a subworkflow, which holds a reference to it.
    """

    def __init__(
        self,
        workflow: Workflow,
        name: str,
        subworkflow: Workflow,
        arguments: Mapping[str, Basic | Reference[Any] | Raw],
        raw_as_parameter: bool = False,
    ):
        """Create a NestedStep.

        Args:
            workflow: outer workflow.
            name: name of the subworkflow.
            subworkflow: inner workflow (subworkflow) itself.
            arguments: arguments provided to the step.
            raw_as_parameter: whether raw-type arguments should be made (outer) workflow parameters.
        """
        self.__subworkflow__ = subworkflow
        base_arguments: dict[str, Basic | Reference[Any] | Raw] = {
            p.name: p for p in subworkflow.find_parameters()
        }
        base_arguments.update(arguments)
        super().__init__(
            workflow=workflow,
            task=subworkflow,
            arguments=base_arguments,
            raw_as_parameter=raw_as_parameter,
        )

    @property
    def subworkflow(self) -> Workflow:
        """Subworkflow that is wrapped."""
        return self.__subworkflow__

    @property
    def return_type(self) -> Any:
        """Take the type of the wrapped function from the target.

        Unwraps and inspects the signature, meaning that the original
        wrapped function _must_ have a typehint for the return value.

        Returns:
            Expected type of the return value.
        """
        result = self.__subworkflow__.result
        if result is None or (isinstance(result, list | tuple) and not result):
            raise RuntimeError("Can only use a subworkflow if the reference exists.")
        return self.__subworkflow__.result_type


class Step(BaseStep):
    """Regular step."""

    ...


class FactoryCall(Step):
    """Call to a factory function."""

    def __init__(
        self,
        workflow: Workflow,
        task: Task | Workflow,
        arguments: Mapping[str, Reference[Any] | Raw],
        raw_as_parameter: bool = False,
    ):
        """Initialize a step.

        Args:
            workflow: `Workflow` that this is tied to.
            task: the lazy-evaluatable function that this wraps.
            arguments: key-value pairs to pass to the function - for a factory call, these _must_ be raw.
            raw_as_parameter: whether to turn any raw-type arguments into workflow parameters (or just keep them as default argument values).
        """
        for _, arg in arguments.items():
            if not is_expr(arg) and not (
                isinstance(arg, ParameterReference) and is_raw_type(arg.__type__)
            ):
                try:
                    arg_name = str(arg)
                except:
                    arg_name = "(unnamed)"
                raise RuntimeError(
                    f"Factories must be constructed with raw types in {arg_name} {type(arg)}"
                )
        super().__init__(
            workflow=workflow,
            task=task,
            arguments=arguments,
            raw_as_parameter=raw_as_parameter,
        )

    @property
    def __name__(self) -> str:
        """Get the name of this factory call."""
        return self.name

    @property
    def __original_name__(self) -> str:
        """Original name of a factory call is just its normal name."""
        return self.name

    @property
    def __default__(self) -> Unset:
        """Dummy default property for use as property."""
        return UnsetType(self.return_type)


class ParameterReference(FieldableMixin, Reference[U], WorkflowComponent):
    """Reference to an individual `Parameter`.

    Allows us to refer to the outputs of a `Parameter` in subsequent `Parameter`
    arguments.

    Attributes:
        parameter: `Parameter` referred to.
        workflow: Related workflow. In this case, as Parameters are generic
            but ParameterReferences are specific, this carries the actual workflow reference.

    Returns:
        Workflow that the referee is related to.
    """

    class ParameterReferenceMetadata(Generic[T]):
        """Holder for attributes of this reference that we do not wish to risk confusing with fieldnames.

        Attributes:
            parameter: the parameter to which this reference refers.
        """

        parameter: Parameter[T]

        def __init__(self, parameter: Parameter[T], typ: type[U] | Unset = UNSET):
            """Initialize the reference.

            Args:
                workflow: `Workflow` that this is tied to.
                parameter: `Parameter` that this refers to.
                typ: type that should override inferred type or None.
                *args: any arguments for other initializers.
                **kwargs: any arguments for other initializers.
            """
            self.parameter = parameter

        @property
        def unique_name(self) -> str:
            """Unique, machine-generated name.

            Normally this will become invisible in output, but it avoids circularity
            as a step that uses this parameter will ask for this when constructing
            its own hash, but we will normally want to use the step's name as part of
            the parameter name to distinguish from other parameters of the same name.
            """
            return self.parameter.__name__

    @property
    def __default__(self) -> U | Unset:
        """Default value of the parameter."""
        default = self._.parameter.default
        if isinstance(default, Unset):
            return default

        for field in self.__field__:
            if isinstance(default, dict) or isinstance(field, int):
                default = default[field]
            else:
                default = getattr(default, field)
        return default

    @property
    def __root_name__(self) -> str:
        """Reference based on the named step.

        May be remapped by the workflow to something nicer
        than the ID.
        """
        return self._.parameter.name

    def __init__(
        self,
        parameter: Parameter[U],
        *args: Any,
        typ: type[U] | None = None,
        **kwargs: Any,
    ):
        """Extract the parameter and type for setup.

        Args:
            parameter: the parameter to reference.
            typ: an overriding type for this reference, or None.
            *args: arguments for other initializers.
            **kwargs: arguments for other initializers.
        """
        chosen_type = typ or parameter.__type__
        self._ = self.ParameterReferenceMetadata(parameter, typ=chosen_type)
        super().__init__(*args, typ=typ, **kwargs)

    def __getitem__(self, attr: str) -> "ParameterReference[U]":
        """Retrieve a field.

        Args:
            attr: attribute to get.

        Returns: a reference to the field within this parameter, possibly nesting if we are already
            referencing a field.
        """
        try:
            return self.find_field(
                field=attr, workflow=self.__workflow__, parameter=self._.parameter
            )
        except AttributeError as exc:
            raise KeyError(
                f"Key not found in {self.__root_name__} ({type(self)}:{self.__type__}): {attr}"
                + (
                    ". This could be because you are trying to iterate/index a reference whose type is not definitely iterable - double check your typehints."
                    if isinstance(attr, int)
                    else ""
                )
            ) from exc

    @property
    def __original_name__(self) -> str:
        """The name of the original parameter, without any field, etc."""
        return self._.parameter.__original_name__

    def __getattr__(self, attr: str) -> "ParameterReference[U] | Any":
        """Retrieve a field.

        Args:
            attr: attribute to get.

        Returns: a reference to the field within this parameter, possibly nesting if we are already
            referencing a field.
        """
        try:
            return self[attr]
        except KeyError as _:
            return super().__getattribute__(attr)

    def __repr__(self) -> str:
        """Hashable reference to the step (and field)."""
        try:
            typ = self.__type__.__name__
        except AttributeError:
            typ = str(self.__type__)
        name = self._.unique_name + self.__field_suffix__
        return f"{typ}|:param:{name}"

    def __hash__(self) -> int:
        """Hash to parameter.

        Returns:
            Unique hash corresponding to the parameter.
        """
        return hash((self._.parameter, self.__field__))

    def __eq__(self, other: object) -> bool:
        """Compare two references.

        We ignore the workflow itself, as equality here is usually
        to test mergeability of the "same" reference in two workflows.

        Returns:
            True if the other parameter reference is materially the same, otherwise False.
        """
        # We are equal to a parameter if we are a direct, fieldless, reference to it.
        return (
            isinstance(other, ParameterReference)
            and self._.parameter == other._.parameter
            and self.__field__ == other.__field__
        )

    def __make_reference__(self, **kwargs: Any) -> "ParameterReference[U]":
        """Get a reference for the same parameter."""
        if "workflow" not in kwargs:
            kwargs["workflow"] = self.__workflow__
        return self._.parameter.make_reference(**kwargs)


class IterableParameterReference(IterableMixin[U], ParameterReference[U]):
    """Iterable form of parameter references."""

    def __iter__(self) -> Generator[Reference[U], None, None]:
        """Iterate over this parameter.

        Returns:
            Get values back for this parameter, normally references of the appropriate type.
        """
        inner, metadata = strip_annotations(self.__type__)
        if (
            metadata
            and "AtRender" in metadata
            and isinstance(self.__default__, Iterable)
        ):
            yield from self.__default__
        else:
            yield from super().__iter__()

    def __inner_iter__(self) -> Generator[Any, None, None]:
        """Iterate over this parameter.

        This is intended for overriding as a convenience way of iterating over a parameter,
        and the yielded values will be ignored, in favour of new references.

        Returns: a generator of any type.
        """
        inner, metadata = strip_annotations(self.__type__)
        if self.__fixed_len__ is not None:
            yield from range(self.__fixed_len__)
        elif metadata and "Fixed" in metadata and isinstance(self.__default__, Sized):
            yield from range(len(self.__default__))
        else:
            while True:
                yield None

    def __len__(self) -> int:
        """If it is possible to get a hard-codeable length from this iterable parameter, do so."""
        inner, metadata = strip_annotations(self.__type__)
        if metadata and "Fixed" in metadata and isinstance(self.__default__, Sized):
            return len(self.__default__)
        return super().__len__()


class StepReference(FieldableMixin, Reference[U]):
    """Reference to an individual `Step`.

    Allows us to refer to the outputs of a `Step` in subsequent `Step`
    arguments.

    Attributes:
        _: metadata wrapping the `Step` referred to.
    """

    step: BaseStep

    class StepReferenceMetadata:
        """Wrapper for any metadata that we would not want to conflict with fieldnames.

        Attributes:
            step: the step being wrapped.
            _typ: the type to return, if overriding the step's own type, or None.
        """

        def __init__(self, step: BaseStep, typ: type[U] | None = None):
            """Initialize the reference.

            Args:
                workflow: `Workflow` that this is tied to.
                step: `Step` that this refers to.
                typ: the type that the step will output.
                field: if provided, a specific field to pull out of an attrs result class.
            """
            self.step = step
            self._typ = typ

        @property
        def return_type(self) -> type:
            """Type of this reference, which may be overridden or may be the step's own type."""
            return self._typ or self.step.return_type

    _: StepReferenceMetadata

    def __init__(
        self, step: BaseStep, *args: Any, typ: type[U] | None = None, **kwargs: Any
    ):
        """Initialize the reference.

        Args:
            workflow: `Workflow` that this is tied to.
            step: `Step` that this refers to.
            typ: the type that the step will output.
            field: if provided, a specific field to pull out of an attrs result class.
            *args: arguments for other initializers.
            **kwargs: arguments for other initializers.
        """
        typ = typ or step.return_type
        self._ = self.StepReferenceMetadata(step, typ=typ)
        super().__init__(*args, typ=typ, **kwargs)

    def __str__(self) -> str:
        """Global description of the reference."""
        return self.__name__

    def __repr__(self) -> str:
        """Hashable reference to the step (and field)."""
        return self._.step.id + self.__field_suffix__

    def __hash__(self) -> int:
        """Hashable value for this workflow."""
        return hash((repr(self), id(self.__workflow__)))

    def __getitem__(self, attr: str) -> "StepReference[Any]":
        """Reference to a field within this result, if possible.

        If the result is an attrs-class or dataclass, this will pull out an individual
        field for use as input to other tasks, or global output of the workflow.

        Args:
            attr: field to pull out.

        Returns:
            Another StepReference specific to the requested field.

        Raises:
            AttributeError: if this field is not present in the dataclass.
            RuntimeError: if this field is not available, or we do not have a structured result.
        """
        try:
            return self.find_field(
                workflow=self.__workflow__, step=self._.step, field=attr
            )
        except AttributeError as exc:
            raise KeyError(
                f"Key not found in {self.__root_name__} ({type(self)}:{self.__type__}): {attr}"
                + (
                    ". This could be because you are trying to iterate/index a reference whose type is not definitely iterable - double check your typehints."
                    if isinstance(attr, int)
                    else ""
                )
            ) from exc

    def __getattr__(self, attr: str) -> "StepReference[U] | Any":
        """Retrieve a field within this workflow."""
        try:
            return self[attr]
        except KeyError as exc:
            try:
                return super().__getattribute__(attr)
            except AttributeError as inner_exc:
                raise inner_exc from exc

    @property
    def __type__(self) -> type:
        """Get the type to which this step reference refers."""
        return self._.return_type

    @property
    def __root_name__(self) -> str:
        """Reference based on the named step.

        May be remapped by the workflow to something nicer
        than the ID.
        """
        return self._.step.name

    @property
    def __workflow__(self) -> WorkflowProtocol:
        """Related workflow.

        Returns:
            Workflow that the referee is related to.
        """
        return self._.step.__workflow__

    @__workflow__.setter
    def __workflow__(self, workflow: Workflow) -> None:
        """Sets related workflow.

        Args:
            workflow: workflow to update the step
        """
        self._.step.set_workflow(workflow)

    def __make_reference__(self, **kwargs: Any) -> "StepReference[U]":
        """Create a new reference for the same step."""
        if "workflow" not in kwargs:
            kwargs["workflow"] = self.__workflow__
        return self._.step.make_reference(**kwargs)


class IterableStepReference(IterableMixin[U], StepReference[U]):
    """Iterable form of a step reference."""

    def __iter__(self) -> Generator[Reference[U], None, None]:
        """Gets a sentinel value for iterating over this step's results.

        Bear in mind that this means an iterable step reference will iterate exactly once,
        and return this Generator. The IteratedGenerator can be used as an iterator itself,
        and will yield an infinite sequence of numbered field references, which renderers can use
        for zipping with a fixed length iterator, or simply prepping fieldnames for serialization.
        """
        # We cast this so that we can treat a step iterator as if it really loops over results.
        yield cast(Reference[U], IteratedGenerator(self))


def is_task(task: Lazy) -> bool:
    """Decide whether this is a task.

    Checks whether the wrapped function has the magic
    attribute `__step_expression__` set to True, which is
    done within task creation.

    Args:
        task: lazy-evaluated value, suspected to be a task.

    Returns:
        True if `task` is indeed a task.
    """
    return isinstance(task, LazyEvaluation)


def expr_to_references(
    expression: Any, remap: Callable[[Any], Any] | None = None
) -> tuple[ExprType, list[Reference[Any] | Parameter[Any]]]:
    """Pull out any references, or other free symbols, from an expression.

    Args:
        expression: normally a reference that can be immediately returned, but may be a sympy expression or a dict/tuple/list/etc. of such.
        remap: a callable to project certain values down before extracting symbols, or None.

    Returns: a pair of the expression with any applied simplifications/standardizations, and the list of References/Parameters found.
    """
    to_check: list[Reference[Any] | Parameter[Any]] = []

    def _to_expr(value: Any) -> ExprType:
        if remap and (res := remap(value)) is not None:
            return _to_expr(res)

        if isinstance(value, Reference):
            to_check.append(value)
            return value

        if value is None:
            return None

        if isinstance(value, Symbol):
            return value
        elif isinstance(value, Basic):
            for sym in value.free_symbols:
                new_sym = _to_expr(sym)
                if new_sym != sym:
                    value = value.subs(sym, new_sym)
            return value

        if is_dataclass(value) or attr_has(value):
            if is_dataclass(value):
                fields = list(dataclass_fields(value))
            else:
                fields = list({field for field in attrs_fields(value)})
            for field in fields:
                if hasattr(value, field.name) and isinstance(
                    (val := getattr(value, field.name)), Reference
                ):
                    setattr(value, field.name, _to_expr(val))
            return value

        # We need to look inside a Raw, but we do not want to lose it if
        # we do not need to.
        retval = value
        if isinstance(value, Raw):
            value = value.value

        if isinstance(value, Mapping):
            dct = {key: _to_expr(val) for key, val in value.items()}
            if dct == value:
                return retval
            # We try to reinstantiate this, but there will be an error otherwise.
            # TODO: we could check this with a protocol, but some care would be needed to ensure all valid
            # cases are covered.
            return value.__class__(dct)  # type: ignore
        elif not isinstance(value, str | bytes) and isinstance(value, Iterable):
            lst = (tuple if isinstance(value, tuple) else list)(
                _to_expr(v) for v in value
            )
            if lst == value:
                return retval
            return lst
        return retval

    expression = _to_expr(expression)

    # if {sym for sym in symbols if not is_raw(sym)} != set(to_check):
    #    print(symbols, to_check, [type(r) for r in symbols])
    #    raise RuntimeError("The only symbols allowed are references (to e.g. step or parameter)")
    return expression, to_check


def unify_workflows(
    expression: Any, base_workflow: Workflow | None, set_only: bool = False
) -> tuple[Basic | None, Workflow | None]:
    """Takes an expression and ensures all of its references exist in the same workflow.

    Args:
        expression: any valid argument to `dewret.workflow.expr_to_references`.
        base_workflow: the desired workflow to align on, or None.
        set_only: whether to bother assimilating all the workflows (False), or to assume that has been done (False).

    Returns: a pair of the (standardized) expression and the workflow that now contains it.
    """
    expression, to_check = expr_to_references(expression)
    if not to_check:
        return expression, base_workflow

    # Build a unified workflow
    collected_workflow = base_workflow or next(iter(to_check)).__workflow__
    # This is a bit of an artifical check. We could rule out non-Workflow WorkflowProtocol implementers and simplify this.
    if not isinstance(collected_workflow, Workflow):
        raise NotImplementedError("Can only unify workflows on the same type: Workflow")
    if not set_only:
        for step_result in to_check:
            new_workflow = step_result.__workflow__
            if not isinstance(new_workflow, Workflow):
                raise NotImplementedError(
                    "Can only unify workflows on the same type: Workflow"
                )
            if (
                collected_workflow != new_workflow
                and collected_workflow
                and new_workflow
            ):
                collected_workflow = Workflow.assimilate(
                    collected_workflow, new_workflow
                )

    # Make sure all the results share it
    for step_result in to_check:
        step_result.__workflow__ = collected_workflow

    return expression, collected_workflow



"""Utility file used to generate recipe.yaml."""

import textwrap
import toml

with open("pyproject.toml") as f:
    pyproject = toml.load(f)

with open("README.md") as f:
    readme_string = f.read()

readme_string = textwrap.indent(readme_string, "    ")

with open("recipe.yaml") as f:
    TEMPLATE = f.read()

deps = pyproject["project"]["dependencies"]
deps_string = ""
for dep in deps:
    deps_string += f"    - {dep}\n"
TEMPLATE = TEMPLATE.replace("__PROJECT_DEPENDENCIES__", deps_string)
TEMPLATE = TEMPLATE.replace("__README__", readme_string)
with open("recipe.yaml", "w") as f:
    f.write(TEMPLATE)



"""Check renderers can be imported live."""

import pytest
from pathlib import Path
from dewret.tasks import construct
from dewret.render import get_render_method

from ._lib.extra import increment, triple_and_one


def test_can_load_render_module() -> None:
    """Checks if we can load a render module."""
    result = triple_and_one(num=increment(num=3))
    workflow = construct(result, simplify_ids=True)
    workflow._name = "Fred"

    frender_py = Path(__file__).parent / "_lib/frender.py"
    render = get_render_method(frender_py)

    assert render(workflow) == {
        "__root__": """
I found a workflow called Fred.
It has 2 steps!
They are:
* Something called increment-1

* A portal called triple_and_one-1 to another workflow,
  whose name is triple_and_one

It probably got made with JUMP=1.0
""",
        "triple_and_one-1": """
I found a workflow called triple_and_one.
It has 3 steps!
They are:
* Something called double-1-1

* Something called sum-1-1

* Something called sum-1-2

It probably got made with JUMP=1.0
""",
    }


def test_can_load_cwl_render_module() -> None:
    """Checks if we can load a render module."""
    result = triple_and_one(num=increment(num=3))
    workflow = construct(result, simplify_ids=True)

    frender_py = Path(__file__).parent.parent / "src/dewret/renderers/cwl.py"
    render = get_render_method(frender_py)
    assert render(workflow) == {
        "__root__": "{'cwlVersion': 1.2, 'class': 'Workflow', 'inputs': {'increment-1-num': {'label': 'num', 'type': 'int', 'default': 3}}, 'outputs': {'out': {'label': 'out', 'type': ['int', 'float'], 'outputSource': 'triple_and_one-1/out'}}, 'steps': {'increment-1': {'run': 'increment', 'in': {'num': {'source': 'increment-1-num'}}, 'out': ['out']}, 'triple_and_one-1': {'run': 'triple_and_one', 'in': {'num': {'source': 'increment-1/out'}}, 'out': ['out']}}}",
        "triple_and_one-1": "{'cwlVersion': 1.2, 'class': 'Workflow', 'inputs': {'num': {'label': 'num', 'type': 'int'}}, 'outputs': {'out': {'label': 'out', 'type': ['int', 'float'], 'outputSource': 'sum-1-1/out'}}, 'steps': {'double-1-1': {'run': 'double', 'in': {'num': {'source': 'num'}}, 'out': ['out']}, 'sum-1-1': {'run': 'sum', 'in': {'left': {'source': 'sum-1-2/out'}, 'right': {'default': 1}}, 'out': ['out']}, 'sum-1-2': {'run': 'sum', 'in': {'left': {'source': 'double-1-1/out'}, 'right': {'source': 'num'}}, 'out': ['out']}}}",
    }


def test_get_correct_import_error_if_unable_to_load_render_module() -> None:
    """Check if the correct import error will be logged if unable to load render module."""
    unfrender_py = Path(__file__).parent / "_lib/unfrender.py"
    with pytest.raises(ModuleNotFoundError) as exc:
        get_render_method(unfrender_py)

    entry = exc.traceback[-1]
    assert (
        Path(entry.path).resolve()
        == (Path(__file__).parent / "_lib" / "unfrender.py").resolve()
    )
    assert entry.relline == 12
    assert "No module named 'extra'" in str(exc.value)

    nonfrender_py = Path(__file__).parent / "_lib/nonfrender.py"
    with pytest.raises(NotImplementedError) as nexc:
        get_render_method(nonfrender_py)

    assert (
        "This render module neither seems to be a structured nor a raw render module"
        in str(nexc.value)
    )



# Copyright 2024- Flax & Teal Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for building renderers.

Provides the routines for calling varied renderers in a standard way, and for
renderers to reuse to build up their own functionality.
"""

import sys
from pathlib import Path
from functools import partial
from typing import TypeVar, Callable, ContextManager, IO, Any, cast
import yaml

from .workflow import Workflow, NestedStep
from .core import (
    RawType,
    RenderCall,
    BaseRenderModule,
    RawRenderModule,
    StructuredRenderModule,
    RenderConfiguration,
)
from .utils import load_module_or_package

T = TypeVar("T")


def structured_to_raw(rendered: RawType, pretty: bool = False) -> str:
    """Serialize a serializable structure to a string.

    Args:
        rendered: a possibly-nested, static basic Python structure.
        pretty: whether to attempt YAML dumping with an indent of 2.

    Returns: YAML/stringified version of the structure.
    """
    if pretty:
        output = yaml.safe_dump(rendered, indent=2)
    else:
        output = str(rendered)
    return output


def get_render_method(
    renderer: Path | RawRenderModule | StructuredRenderModule, pretty: bool = False
) -> RenderCall:
    """Create a ready-made callable to render the workflow that is appropriate for the renderer module.

    Args:
        renderer: a module or path to a module.
        pretty: whether the renderer should attempt to YAML-format the output (if relevant).

    Returns: a callable with a consistent interface, regardless of the renderer type.
    """
    render_module: BaseRenderModule
    if isinstance(renderer, Path):
        # Attempt to load renderer as package, falling back to a single module otherwise.
        # This enables relative imports in renderers and therefore the ability to modularize.
        module = load_module_or_package("__renderer__", renderer)
        sys.modules["__renderer_mod__"] = module
        render_module = cast(BaseRenderModule, module)
    else:
        render_module = renderer

    if isinstance(render_module, RawRenderModule):
        return render_module.render_raw
    elif isinstance(render_module, (StructuredRenderModule)):

        def _render(
            workflow: Workflow,
            render_module: StructuredRenderModule,
            pretty: bool = False,
            **kwargs: RenderConfiguration,
        ) -> dict[str, str]:
            rendered = render_module.render(workflow, **kwargs)
            return {
                key: structured_to_raw(value, pretty=pretty)
                for key, value in rendered.items()
            }

        return cast(
            RenderCall, partial(_render, render_module=render_module, pretty=pretty)
        )

    raise NotImplementedError(
        "This render module neither seems to be a structured nor a raw render module."
    )


def write_rendered_output(
    rendered: dict[str, str] | dict[str, RawType],
    output: str,
    opener: Callable[[str, str], ContextManager[IO[Any]]],
) -> None:
    """Utility function to handle writing rendered output to file or stdout."""
    if len(rendered) == 1:
        with opener("", "w") as output_f:
            output_f.write(rendered["__root__"])
    elif "%" in output:
        for key, value in rendered.items():
            if key == "__root__":
                key = "ROOT"
            with opener(key, "w") as output_f:
                output_f.write(value)
    else:
        with opener("ROOT", "w") as output_f:
            output_f.write(rendered["__root__"])
        del rendered["__root__"]
        for key, value in rendered.items():
            with opener(key, "a") as output_f:
                output_f.write("\n---\n")
                output_f.write(value)


def base_render(workflow: Workflow, build_cb: Callable[[Workflow], T]) -> dict[str, T]:
    """Render to a dict-like structure.

    Args:
        workflow: workflow to evaluate result.
        build_cb: a callback to call for each workflow found.

    Returns:
        Reduced form as a native Python dict structure for
        serialization.
    """
    primary_workflow = build_cb(workflow)
    subworkflows = {}
    for step in workflow.indexed_steps.values():
        if isinstance(step, NestedStep):
            nested_subworkflows = base_render(step.subworkflow, build_cb)
            subworkflows.update(nested_subworkflows)
            subworkflows[step.name] = nested_subworkflows["__root__"]
    subworkflows["__root__"] = primary_workflow
    return subworkflows



"""Supporting tasks for complex workflow.

Illustrates the ability to modularize code.
"""

from dewret.tasks import task

JUMP: int = 10


@task()
def increase(num: int) -> int:
    """Add globally-configured integer JUMP to a number."""
    return num + JUMP


@task()
def increment(num: int) -> int:
    """Increment an integer."""
    return num + 1


@task()
def double(num: int) -> int:
    """Double an integer."""
    return 2 * num


@task()
def mod10(num: int) -> int:
    """Calculate supplied integer modulo 10."""
    return num % 10


@task()
def sum(left: int, right: int) -> int:
    """Add two integers."""
    return left + right



# Insert Python copyright
# This file: PSL 2.0

import builtins
import dis
from inspect import (
    ClosureVars,
    ismethod,
    isfunction,
    ismodule,
)
from typing import Callable, Any


def getclosurevars(func: Callable[..., Any]) -> ClosureVars:
    if ismethod(func):
        func = func.__func__

    if not isfunction(func):
        raise TypeError("{!r} is not a Python function".format(func))

    code = func.__code__
    # Nonlocal references are named in co_freevars and resolved
    # by looking them up in __closure__ by positional index
    if func.__closure__ is None:
        nonlocal_vars = {}
    else:
        nonlocal_vars = {
            var: cell.cell_contents
            for var, cell in zip(code.co_freevars, func.__closure__, strict=False)
        }

    # Global and builtin references are named in co_names and resolved
    # by looking them up in __globals__ or __builtins__
    global_ns = func.__globals__
    builtin_ns = global_ns.get("__builtins__", builtins.__dict__)
    if ismodule(builtin_ns):
        builtin_ns = builtin_ns.__dict__
    global_vars = {}
    builtin_vars = {}
    unbound_names = set()
    global_names = set()
    stackm1 = None
    imports: dict[str, str] = {}
    for instruction in dis.get_instructions(code):
        opname = instruction.opname
        name = instruction.argval
        # TODO: this does not cover the possibility of a subsequent
        # variable shadowing an inline import
        if opname == "LOAD_ATTR" and stackm1 and (fullname := imports.get(stackm1)):
            unbound_names.add(f"{stackm1}.{name}:{fullname}.{name}")
            stackm1 = None
        elif opname == "IMPORT_NAME":
            # TODO: deal with STACK[-2]
            # Strictly, these only enter the namespace with a
            # subsequent STORE_FAST, present in normal circumstances.
            if stackm1 is not None:
                fullname = f"{stackm1}.{name}"
            else:
                fullname = name
            imports[name] = fullname
            stackm1 = name
        elif opname == "IMPORT_FROM":
            fullname = f"{stackm1}.{name}"
            unbound_names.add(f"{name}:{fullname}")
            imports[name] = fullname
            stackm1 = name
        elif opname == "LOAD_FAST":
            stackm1 = name
        elif opname == "LOAD_GLOBAL":
            global_names.add(name)
    for name in global_names:
        try:
            global_vars[name] = global_ns[name]
        except KeyError:
            try:
                builtin_vars[name] = builtin_ns[name]
            except KeyError:
                unbound_names.add(name)

    return ClosureVars(nonlocal_vars, global_vars, builtin_vars, unbound_names)



# Copyright 2014 Flax & Teal Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""CWL Renderer.

Outputs a [Common Workflow Language](https://www.commonwl.org/) representation of the
current workflow.
"""

from attrs import define, has as attrs_has, fields as attrs_fields, AttrsInstance
from dataclasses import is_dataclass, fields as dataclass_fields
from collections.abc import Mapping
from typing import (
    TypedDict,
    NotRequired,
    get_origin,
    get_args,
    cast,
    Any,
    Unpack,
    Iterable,
    Type,
    TypeVar,
)
from types import UnionType
from inspect import isclass
from sympy import Basic, Tuple, Dict, jscode, Symbol

from dewret.core import (
    Raw,
    RawType,
    FirmType,
)
from dewret.workflow import (
    FactoryCall,
    Workflow,
    BaseStep,
    StepReference,
    ParameterReference,
    expr_to_references,
)
from dewret.utils import (
    crawl_raw,
    DataclassProtocol,
    firm_to_raw,
    flatten_if_set,
    Unset,
)
from dewret.render import base_render
from dewret.core import Reference, get_render_configuration, set_render_configuration


class CommandInputSchema(TypedDict):
    """Structure for referring to a raw type in CWL.

    Encompasses several CWL types. In future, it may be best to
    use _cwltool_ or another library for these basic structures.

    Attributes:
        type: CWL type of this input.
        label: name to show for this input.
        fields: (for `record`) individual fields in a dict-like structure.
        items: (for `array`) type that each field will have.
    """

    type: "InputSchemaType"
    label: str
    fields: NotRequired[dict[str, "CommandInputSchema"]]
    items: NotRequired["InputSchemaType"]
    default: NotRequired[RawType]


T = TypeVar("T")

InputSchemaType = (
    str
    | CommandInputSchema
    | list[str]
    | list["InputSchemaType"]
    | dict[str, "str | InputSchemaType"]
)


def render_expression(ref: Any) -> "ReferenceDefinition":
    """Turn a rich (sympy) expression into a CWL JS expression.

    Args:
        ref: a structure whose elements are all string-renderable or sympy Basic.

    Returns: a ReferenceDefinition containing a string representation of the expression in the form `$(...)`.
    """

    def _render(ref: Any) -> Basic | RawType:
        if not isinstance(ref, Basic):
            if isinstance(ref, Mapping):
                ref = Dict({key: _render(val) for key, val in ref.items()})
            elif not isinstance(ref, str | bytes) and isinstance(ref, Iterable):
                ref = Tuple(*(_render(val) for val in ref))
        return ref

    expr = _render(ref)
    if isinstance(expr, Basic):
        values = list(expr.free_symbols)
        step_syms = [sym for sym in expr.free_symbols if isinstance(sym, StepReference)]
        param_syms = [
            sym for sym in expr.free_symbols if isinstance(sym, ParameterReference)
        ]

        if set(values) != set(step_syms) | set(param_syms):
            raise NotImplementedError(
                f"Can only build expressions for step results and param results: {ref}"
            )

        if len(step_syms) > 1:
            raise NotImplementedError(
                f"Can only create expressions with 1 step reference: {ref}"
            )
        if not (step_syms or param_syms):
            ...
        if values == [ref]:
            if isinstance(ref, StepReference):
                return ReferenceDefinition(source=to_name(ref), value_from=None)
            else:
                return ReferenceDefinition(source=ref.name, value_from=None)
        source = None
        for ref in values:
            if isinstance(ref, StepReference):
                field = with_field(ref)
                parts = field.split("/")
                base = f"/{parts[0]}" if parts and parts[0] else ""
                if len(parts) > 1:
                    expr = expr.subs(ref, f"self.{'.'.join(parts[1:])}")
                else:
                    expr = expr.subs(ref, "self")
                source = f"{ref.__root_name__}{base}"
            else:
                expr = expr.subs(ref, Symbol(f"inputs.{ref.name}"))
        return ReferenceDefinition(
            source=source, value_from=f"$({jscode(_render(expr))})"
        )
    return ReferenceDefinition(source=str(expr), value_from=None)


class CWLRendererConfiguration(TypedDict):
    """Configuration for the renderer.

    Attributes:
        allow_complex_types: can input/output types be other than raw?
        factories_as_params: should factories be treated as input or steps?
    """

    allow_complex_types: NotRequired[bool]
    factories_as_params: NotRequired[bool]


def default_config() -> CWLRendererConfiguration:
    """Default configuration for this renderer.

    This is a hook-like call to give a configuration dict that this renderer
    will respect, and sets any necessary default values.

    Returns: a dict with (preferably) raw type structures to enable easy setting
        from YAML/JSON.
    """
    return {
        "allow_complex_types": False,
        "factories_as_params": False,
    }


def with_type(result: Any) -> type | Any:
    """Get a Python type from a value.

    Does so either by using its `__type__` field (for example, for References)
    or if unavailable, using `type()`.

    Returns: a Python type.
    """
    if hasattr(result, "__type__"):
        return result.__type__
    return type(result)


def with_field(result: Any) -> str:
    """Get a string representing any 'field' suffix of a value.

    This only makes sense in the context of a Reference, which can represent
    a deep reference with a known variable (parameter or step result, say) using
    its `__field__` attribute. Defaults to `"out"` as this produces compliant CWL
    where every output has a "fieldname".

    Returns: a string representation of the field portion of the passed value or `"out"`.
    """
    if hasattr(result, "__field__") and result.__field__:
        return str(result.__field_str__)
    else:
        return "out"


def to_name(result: Reference[Any]) -> str:
    """Take a reference and get a name representing it.

    The primary purpose of this method is to deal with the case where a reference is to the
    whole result, as we always put this into an imagined `out` field for CWL consistency.

    Returns: the name of the reference, including any field portion, appending an `"out"` fieldname if none.
    """
    if (
        hasattr(result, "__field__")
        and not result.__field__
        and isinstance(result, StepReference)
    ):
        return f"{result.__name__}/out"
    return result.__name__


@define
class ReferenceDefinition:
    """CWL-renderable internal reference.

    Normally points to a value or a step.
    """

    source: str | None
    value_from: str | None

    @classmethod
    def from_reference(cls, ref: Reference[Any]) -> "ReferenceDefinition":
        """Build from a `Reference`.

        Converts a `dewret.workflow.Reference` into a CWL-rendering object.

        Args:
            ref: reference to convert.
        """
        return render_expression(ref)

    def render(self) -> dict[str, RawType]:
        """Render to a dict-like structure.

        Returns:
            Reduced form as a native Python dict structure for
            serialization.
        """
        representation: dict[str, RawType] = {}
        if self.source is not None:
            representation["source"] = self.source
        if self.value_from is not None:
            representation["valueFrom"] = self.value_from
        return representation


@define
class StepDefinition:
    """CWL-renderable step.

    Coerces the dewret structure of a step into that
    needed for valid CWL.

    Attributes:
        name: identifier to call this step by.
        run: task to execute for this step.
        in_: inputs from values or other steps.
    """

    name: str
    run: str
    out: dict[str, "CommandInputSchema"] | list[str]
    in_: Mapping[str, ReferenceDefinition | Raw]

    @classmethod
    def from_step(cls, step: BaseStep) -> "StepDefinition":
        """Build from a `BaseStep`.

        Converts a `dewret.workflow.Step` into a CWL-rendering object.

        Args:
            step: step to convert.
        """
        out: list[str] | dict[str, "CommandInputSchema"]
        if attrs_has(step.return_type) or (
            is_dataclass(step.return_type) and isclass(step.return_type)
        ):
            out = to_output_schema("out", step.return_type)["fields"]
        else:
            out = ["out"]
        return cls(
            name=step.name,
            run=step.task.name,
            out=out,
            in_={
                key: (
                    ReferenceDefinition.from_reference(param)
                    if isinstance(param, Reference)
                    else param
                )
                for key, param in step.arguments.items()
            },
        )

    def render(self) -> dict[str, RawType]:
        """Render to a dict-like structure.

        Returns:
            Reduced form as a native Python dict structure for
            serialization.
        """
        return {
            "run": self.run,
            "in": {
                key: (
                    ref.render()
                    if isinstance(ref, ReferenceDefinition)
                    else render_expression(ref).render()
                    if isinstance(ref, Basic)
                    else {"default": firm_to_raw(ref.value)}
                    if hasattr(ref, "value")
                    else render_expression(ref).render()
                )
                for key, ref in self.in_.items()
            },
            "out": crawl_raw(self.out),
        }


def cwl_type_from_value(label: str, val: RawType | Unset) -> CommandInputSchema:
    """Find a CWL type for a given (possibly Unset) value.

    Args:
        label: the label for the variable being checked to prefill the input def and improve debugging info.
        val: a raw Python variable or an unset variable.

    Returns:
        Input schema type.
    """
    if val is not None and hasattr(val, "__type__"):
        raw_type = val.__type__
    else:
        raw_type = type(val)

    return to_cwl_type(label, raw_type)


def to_cwl_type(label: str, typ: Type[T]) -> CommandInputSchema:
    """Map Python types to CWL types.

    Args:
        label: the label for the variable being checked to prefill the input def and improve debugging info.
        typ: a Python basic type.

    Returns:
        CWL specification type dict.
    """
    typ_dict: CommandInputSchema = {"label": label, "type": ""}
    base: Any | None = typ
    args = get_args(typ)
    if args:
        base = get_origin(typ)

    if base == type(None):
        typ_dict["type"] = "null"
    elif base == int:
        typ_dict["type"] = "int"
    elif base == bool:
        typ_dict["type"] = "boolean"
    elif base == dict or (isinstance(base, type) and attrs_has(base)):
        typ_dict["type"] = "record"
    elif base == float:
        typ_dict["type"] = "float"
    elif base == str:
        typ_dict["type"] = "string"
    elif base == bytes:
        typ_dict["type"] = "bytes"
    elif isinstance(typ, UnionType):
        typ_dict.update(
            {"type": tuple(to_cwl_type(label, item)["type"] for item in args)}
        )
    elif isclass(base) and issubclass(base, Iterable):
        try:
            if len(args) > 1:
                typ_dict.update(
                    {
                        "type": "array",
                        "items": [to_cwl_type(label, t)["type"] for t in args],
                    }
                )
            elif len(args) == 1:
                typ_dict.update(
                    {"type": "array", "items": to_cwl_type(label, args[0])["type"]}
                )
            else:
                typ_dict["type"] = "array"
        except IndexError as err:
            raise TypeError(
                f"Cannot render complex type ({typ}) to CWL for {label}, have you enabled allow_complex_types configuration?"
            ) from err
    elif get_render_configuration("allow_complex_types"):
        typ_dict["type"] = typ if isinstance(typ, str) else typ.__name__
    else:
        raise TypeError(f"Cannot render type ({typ}) to CWL for {label}")
    return typ_dict


class CommandOutputSchema(CommandInputSchema):
    """Structure for referring to an output in CWL.

    As a simplification, this is an input schema with an extra
    `outputSource` field.

    Attributes:
        outputSource: step result to use for this output.
    """

    outputSource: NotRequired[str]
    expression: NotRequired[str]
    source: NotRequired[list[str]]


def raw_to_command_input_schema(label: str, value: RawType | Unset) -> InputSchemaType:
    """Infer the CWL input structure for this value.

    Inspects the value, to work out an appropriate structure
    describing it in CWL.

    Args:
        label: name of the variable.
        value: basic-typed variable from which to build structure.

    Returns:
        Structure used to define (possibly compound) basic types for input.
    """
    if isinstance(value, dict) or isinstance(value, list):
        return _raw_to_command_input_schema_internal(label, value)
    else:
        return cwl_type_from_value(label, value)


def to_output_schema(
    label: str,
    typ: type[RawType | AttrsInstance | DataclassProtocol],
    output_source: str | None = None,
) -> CommandOutputSchema:
    """Turn a step's output into an output schema.

    Takes a source, type and label and provides a description for CWL.

    Args:
        label: name of this field.
        typ: either a basic type, compound of basic types, or a TypedDict representing a pre-defined result structure.
        output_source: if provided, a CWL step result reference to input here.

    Returns:
        CWL CommandOutputSchema-like structure for embedding into an `outputs` block
    """
    fields = None
    if attrs_has(typ):
        fields = {
            str(field.name): cast(
                CommandInputSchema, to_output_schema(field.name, field.type)
            )
            for field in attrs_fields(typ)
        }
    elif is_dataclass(typ):
        fields = {}
        for field in dataclass_fields(typ):
            # Ideally would raise a type error if the dataclass fields are not valid
            # for to_output_schema, but given that we have no simple, 3.11-compatible
            # way of accepting things like generics and forward references, this turns
            # out to be non-trivial.
            #  raise TypeError("Types of fields in results must also be valid result-types themselves (string-defined types not currently allowed)")
            fields[str(field.name)] = cast(
                CommandInputSchema,
                to_output_schema(field.name, field.type),  # type: ignore
            )
    if fields:
        output = CommandOutputSchema(
            type="record",
            label=label,
            fields=fields,
        )
    else:
        # TODO: this complains because NotRequired keys are never present,
        # but that does not seem like a problem here - likely a better solution.
        output = CommandOutputSchema(
            **to_cwl_type(label, typ)  # type: ignore
        )
    if output_source is not None:
        output["outputSource"] = output_source
    return output


def _raw_to_command_input_schema_internal(
    label: str, value: RawType | Unset
) -> CommandInputSchema:
    structure: CommandInputSchema = cwl_type_from_value(label, value)
    if isinstance(value, dict):
        structure["fields"] = {
            key: _raw_to_command_input_schema_internal(key, val)
            for key, val in value.items()
        }
    elif isinstance(value, list):
        typeset = set(get_args(value))
        if not typeset:
            typeset = {
                item.__type__
                if item is not None and hasattr(item, "__type__")
                else type(item)
                for item in value
            }
        if len(typeset) != 1:
            raise RuntimeError(
                "For CWL, an input array must have a consistent type, "
                "and we need at least one element to infer it, or an explicit typehint."
            )
        structure["items"] = to_cwl_type(label, typeset.pop())["type"]
    elif not isinstance(value, Unset):
        structure["default"] = firm_to_raw(value)
    return structure


@define
class InputsDefinition:
    """CWL-renderable representation of an input parameter block.

    Turns dewret results into a CWL input block.

    Attributes:
        input: sequence of results from a workflow.
    """

    inputs: dict[str, "CommandInputParameter"]

    @define
    class CommandInputParameter:
        """CWL-renderable reference to a specific input.

        Attributes:
            type: type of variable
            name: fully-qualified name of the input.
        """

        type: InputSchemaType
        default: RawType | Unset
        label: str

    @classmethod
    def from_parameters(
        cls, parameters: list[ParameterReference[Any] | FactoryCall]
    ) -> "InputsDefinition":
        """Takes a list of parameters into a CWL structure.

        Uses the parameters to fill out the necessary input fields.

        Returns:
            CWL-like structure representing all workflow outputs.
        """
        parameters_dedup = {
            p._.parameter for p in parameters if isinstance(p, ParameterReference)
        }
        parameters = list(parameters_dedup) + [
            p for p in parameters if not isinstance(p, ParameterReference)
        ]
        return cls(
            inputs={
                input.name: cls.CommandInputParameter(
                    label=input.__name__,
                    default=(default := flatten_if_set(input.__default__)),
                    type=raw_to_command_input_schema(
                        label=input.__original_name__, value=default
                    ),
                )
                for input in parameters
            }
        )

    def render(self) -> dict[str, RawType]:
        """Render to a dict-like structure.

        Returns:
            Reduced form as a native Python dict structure for
            serialization.
        """
        result: dict[str, RawType] = {}
        for key, input in self.inputs.items():
            # Would rather not cast, but CommandInputSchema is dict[RawType]
            # by construction, where type is seen as a TypedDict subclass.
            item = firm_to_raw(cast(FirmType, input.type))
            if isinstance(item, dict) and not isinstance(input.default, Unset):
                item["default"] = firm_to_raw(input.default)
            result[key] = item
        return result


@define
class OutputsDefinition:
    """CWL-renderable set of workflow outputs.

    Turns dewret results into a CWL output block.

    Attributes:
        outputs: sequence of results from a workflow.
    """

    outputs: (
        dict[str, "CommandOutputSchema"]
        | list["CommandOutputSchema"]
        | CommandOutputSchema
    )

    @classmethod
    def from_results(
        cls,
        results: dict[str, StepReference[Any]]
        | list[StepReference[Any]]
        | tuple[StepReference[Any], ...],
    ) -> "OutputsDefinition":
        """Takes a mapping of results into a CWL structure.

        Pulls the result type from the signature, ultimately, if possible.

        Returns:
            CWL-like structure representing all workflow outputs.
        """

        def _build_results(result: Any) -> RawType:
            if isinstance(result, Reference):
                # TODO: need to work out how to tell mypy that a TypedDict is also dict[str, RawType]
                return to_output_schema(  # type: ignore
                    with_field(result), with_type(result), output_source=to_name(result)
                )
            results = result
            return (
                [_build_results(result) for result in results]
                if isinstance(results, list | tuple | Tuple)
                else {key: _build_results(result) for key, result in results.items()}
            )

        try:
            # TODO: sort out this nested type building.
            return cls(outputs=_build_results(results))  # type: ignore
        except AttributeError:
            expr, references = expr_to_references(results)
            reference_names = sorted(
                {
                    str(ref._.parameter)
                    if isinstance(ref, ParameterReference)
                    else str(ref._.step)
                    for ref in references
                }
            )
            return cls(
                outputs={
                    "out": {
                        "type": "float",  # WARNING: we assume any arithmetic expression returns a float.
                        "label": "out",
                        "expression": str(expr),
                        "source": reference_names,
                    }
                }
            )

    def render(self) -> dict[str, RawType] | list[RawType]:
        """Render to a dict-like structure.

        Returns:
            Reduced form as a native Python dict structure for
            serialization.
        """
        return (
            [crawl_raw(output) for output in self.outputs]
            if isinstance(self.outputs, list)
            else {key: crawl_raw(output) for key, output in self.outputs.items()}
        )


@define
class WorkflowDefinition:
    """CWL-renderable workflow.

    Coerces the dewret structure of a workflow into that
    needed for valid CWL.

    Attributes:
        steps: sequence of steps in the workflow.
    """

    steps: list[StepDefinition]
    inputs: InputsDefinition
    outputs: OutputsDefinition
    name: None | str

    @classmethod
    def from_workflow(
        cls, workflow: Workflow, name: None | str = None
    ) -> "WorkflowDefinition":
        """Build from a `Workflow`.

        Converts a `dewret.workflow.Workflow` into a CWL-rendering object.

        Args:
            workflow: workflow to convert.
            name: name of this workflow, if it should have one.
        """
        parameters: list[ParameterReference[Any] | FactoryCall] = list(
            workflow.find_parameters(
                include_factory_calls=not get_render_configuration(
                    "factories_as_params"
                )
            )
        )
        if get_render_configuration("factories_as_params"):
            parameters += list(workflow.find_factories().values())
        return cls(
            steps=[
                StepDefinition.from_step(step)
                for step in workflow.indexed_steps.values()
                if not (
                    isinstance(step, FactoryCall)
                    and get_render_configuration("factories_as_params")
                )
            ],
            inputs=InputsDefinition.from_parameters(parameters),
            outputs=OutputsDefinition.from_results(
                workflow.result
                if isinstance(workflow.result, list | tuple | Tuple)
                else {with_field(workflow.result): workflow.result}
                if workflow.has_result and workflow.result is not None
                else {}
            ),
            name=name,
        )

    def render(self) -> dict[str, RawType]:
        """Render to a dict-like structure.

        Returns:
            Reduced form as a native Python dict structure for
            serialization.
        """
        return {
            "cwlVersion": 1.2,
            "class": "Workflow",
            "inputs": self.inputs.render(),
            "outputs": self.outputs.render(),
            "steps": {step.name: step.render() for step in self.steps},
        }


def render(
    workflow: Workflow, **kwargs: Unpack[CWLRendererConfiguration]
) -> dict[str, dict[str, RawType]]:
    """Render to a dict-like structure.

    Args:
        workflow: workflow to evaluate result.
        **kwargs: additional configuration arguments - these should match CWLRendererConfiguration.

    Returns:
        Reduced form as a native Python dict structure for
        serialization.
    """
    # TODO: Again, convincing mypy that a TypedDict has RawType values.
    with set_render_configuration(kwargs):  # type: ignore
        rendered = base_render(
            workflow,
            lambda workflow: WorkflowDefinition.from_workflow(workflow).render(),
        )
    return rendered



"""Complex workflow.

Useful as an example of a workflow with a nested task.

```sh
$ python -m dewret workflow_complex.py --pretty nested_workflow
```
"""

from dewret.tasks import workflow
from workflow_tasks import sum, double, increase

STARTING_NUMBER: int = 23


@workflow()
def nested_workflow() -> int | float:
    """Creates a complex workflow with a nested task.

    Workflow Steps:
    1. **Increase**: The starting number (`STARTING_NUMBER`) is incremented by 1 using the `increase` task.
    2. **Double**: The result from the first step is then doubled using the `double` task.
    3. **Increase Again**: Separately, the number 17 is incremented twice using the `increase` task.
    4. **Sum**: Finally, the results of the two branches (left and right) are summed together using the `sum` task.

    Returns:
    - `int | float`: The result of summing the doubled and increased values, which may be an integer or a float depending on the operations.

    """
    left = double(num=increase(num=STARTING_NUMBER))
    right = increase(num=increase(num=17))
    return sum(left=left, right=right)



"""Testing example renderer.

Correctly fails to import to show a broken module being handled.
"""

from typing import Unpack, TypedDict
from dewret.workflow import Workflow

# This lacking a relative import, while extra itself
# uses one is what breaks the module. It cannot be both
# a package and not-a-package. This is importable by
# adding a . before extra.
from extra import JUMP  # type: ignore


class UnfrenderRendererConfiguration(TypedDict):
    allow_complex_types: bool


def default_config() -> UnfrenderRendererConfiguration:
    return {"allow_complex_types": True}


def render_raw(
    workflow: Workflow, **kwargs: Unpack[UnfrenderRendererConfiguration]
) -> dict[str, str]:
    return {"JUMP": str(JUMP)}



"""Quickstart workflow.

Useful as an example of a simple workflow.

```sh
$ python -m dewret workflow_qs.py --pretty increment num:3 --backend DASK
```
"""

from dewret.tasks import task


@task()
def increment(num: int) -> int:
    """Add 1 to a number."""
    return num + 1
